@@ -9,8 +9,8 @@
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
- * Copyright (c) 2007-2009 University of Houston. All rights reserved.
- * Copyright (c) 2007-2008 Cisco, Inc. All rights reserved.
+ * Copyright (c) 2007-2010 University of Houston. All rights reserved.
+ * Copyright (c) 2007-2008 Cisco Systems, Inc. All rights reserved.
  * Copyright (c) 2009      Sun Microsystems, Inc.  All rights reserved.
  * $COPYRIGHT$
  * 
@@ -27,12 +27,12 @@
 
 #include "opal/dss/dss.h"
 #include "orte/util/name_fns.h"
+#include "orte/mca/rml/rml_types.h"
 
 #include "ompi/proc/proc.h"
 #include "opal/threads/mutex.h"
 #include "opal/util/bit_ops.h"
-#include "orte/util/show_help.h"
-#include "opal/util/convert.h"
+#include "opal/util/output.h"
 #include "ompi/mca/topo/topo.h"
 #include "ompi/mca/topo/base/base.h"
 #include "ompi/mca/dpm/dpm.h"
@@ -350,7 +350,13 @@ int ompi_comm_create ( ompi_communicator_t *comm, ompi_group_t *group,
              newcomp->c_contextid, comm->c_contextid );
 
     /* Activate the communicator and init coll-component */
-    rc = ompi_comm_activate( &newcomp, 1 );  /* new communicator */ 
+    rc = ompi_comm_activate( &newcomp, /* new communicator */ 
+	                     comm, 
+	                     NULL, 
+	                     NULL, 
+			     NULL,
+	                     mode, 
+	                     -1 );  
     if ( OMPI_SUCCESS != rc ) {
         goto exit;
     }
@@ -572,18 +578,23 @@ int ompi_comm_split ( ompi_communicator_t* comm, int color, int key,
     snprintf(newcomp->c_name, MPI_MAX_OBJECT_NAME, "MPI COMMUNICATOR %d SPLIT FROM %d", 
              newcomp->c_contextid, comm->c_contextid );
 
+    /* set the rank to MPI_UNDEFINED. This prevents in comm_activate
+     * the collective module selection for a communicator that will
+     * be freed anyway.
+     */
+    if ( MPI_UNDEFINED == color ) {
+	newcomp->c_local_group->grp_my_rank = MPI_UNDEFINED;
+    }
+
+
     /* Activate the communicator and init coll-component */
-    if ( MPI_UNDEFINED != color ) {
-	rc = ompi_comm_activate( &newcomp, 1 );  /* new communicator */ 
-    }
-    else {
-	/* 
-        ** this should prevent creating the subcommunicators
-        ** of the hierarch module for the cases where we
-        ** know that they will be freed anyway (e.g. color = MPI_UNDEFINED
-	*/
-	rc = ompi_comm_activate( &newcomp, 0 );  /* new communicator */ 
-    }
+    rc = ompi_comm_activate( &newcomp, /* new communicator */ 
+			     comm, 
+			     NULL,
+			     NULL, 
+			     NULL, 
+			     mode, 
+			     -1 );  
     if ( OMPI_SUCCESS != rc ) {
         goto exit;
     }
@@ -674,7 +685,13 @@ int ompi_comm_dup ( ompi_communicator_t * comm, ompi_communicator_t **newcomm )
              newcomp->c_contextid, comm->c_contextid );
 
     /* activate communicator and init coll-module */
-    rc = ompi_comm_activate( &newcomp, 1 );  /* new communicator */ 
+    rc = ompi_comm_activate( &newcomp, /* new communicator */ 
+			     comp,
+	                     NULL, 
+	                     NULL, 
+	                     NULL, 
+	                     mode, 
+	                     -1 );  
     if ( OMPI_SUCCESS != rc ) {
         return rc;
     }
@@ -934,6 +951,9 @@ static int ompi_comm_allgather_emulate_intra( void *inbuf, int incount,
 int ompi_comm_free ( ompi_communicator_t **comm )
 {
     int ret;
+    int cid = (*comm)->c_contextid;
+    int is_extra_retain = OMPI_COMM_IS_EXTRA_RETAIN(*comm);
+
 
     /* Release attributes.  We do this now instead of during the
        communicator destructor for 2 reasons:
@@ -979,6 +999,30 @@ int ompi_comm_free ( ompi_communicator_t **comm )
     }
     OBJ_RELEASE ( (*comm) );
 
+    if ( is_extra_retain) {
+        /* This communicator has been marked as an "extra retain"
+         * communicator. This can happen if a communicator creates
+         * 'dependent' subcommunicators (e.g. for inter
+         * communicators or when using hierarch collective
+         * module *and* the cid of the dependent communicator
+         * turned out to be lower than of the parent one.
+         * In that case, the reference counter has been increased
+         * by one more, in order to handle the scenario,
+         * that the user did not free the communicator.
+         * Note, that if we enter this routine, we can
+         * decrease the counter by one more therefore. However,
+         * in ompi_comm_finalize, we only used OBJ_RELEASE instead
+         * of ompi_comm_free(), and the increased reference counter
+         * makes sure that the pointer to the dependent communicator
+         * still contains a valid object.
+         */
+        ompi_communicator_t *tmpcomm = (ompi_communicator_t *) opal_pointer_array_get_item(&ompi_mpi_communicators, cid);
+        if ( NULL != tmpcomm ){
+            OBJ_RELEASE (tmpcomm);
+        }
+    }
+
+
     *comm = MPI_COMM_NULL;
     return OMPI_SUCCESS;
 }
@@ -1511,12 +1555,15 @@ int ompi_topo_create (ompi_communicator_t *old_comm,
         return ret;
     }
 
-    if ( MPI_UNDEFINED != new_rank ) {
-	ret = ompi_comm_activate( &new_comm, 1 );  /* new communicator */
-    }
-    else {
-	ret = ompi_comm_activate( &new_comm, 0 );  /* new communicator */
-    }
+    ret = ompi_comm_activate( &new_comm,  /* new communicator */
+                              old_comm,     /* old comm */
+                              NULL,     /* bridge comm */
+                              NULL,     /* local leader */
+                              NULL,     /* remote_leader */
+                              OMPI_COMM_CID_INTRA,   /* mode */
+                             -1 );     /* send first, doesn't matter */
+
+
     if (OMPI_SUCCESS != ret) {
 	/* something wrong happened during setting the communicator */
 	*comm_topo = new_comm;
