@@ -2,13 +2,14 @@
  * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2005 The University of Tennessee and The University
+ * Copyright (c) 2004-2009 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2007-2008 Cisco Systems, Inc.  All rights reserved.
  * $COPYRIGHT$
  * 
  * Additional copyrights may follow
@@ -27,16 +28,23 @@
 
 #include "ompi_config.h"
 
-#if defined(c_plusplus) || defined(__cplusplus)
-extern "C" {
-#endif
-    /**
-     * Wait for a TotalView-like debugger if asked.
-     */
-    OMPI_DECLSPEC void ompi_wait_for_totalview(void);
+BEGIN_C_DECLS
 
-#if defined(c_plusplus) || defined(__cplusplus)
-}
-#endif
+    /**
+     * Wait for a debugger if asked.
+     */
+    extern void ompi_wait_for_debugger(void);
+
+    /**
+     * Notify a debugger that we're about to abort
+     */
+    extern void ompi_debugger_notify_abort(char *string);
+
+    /**
+     * Breakpoint function for parallel debuggers.
+     */
+    OMPI_DECLSPEC void *MPIR_Breakpoint(void);
+
+END_C_DECLS
 
 #endif /* OMPI_DEBUGGERS_H */
