@@ -9,10 +9,11 @@
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
- * Copyright (c) 2006-2008 Cisco, Inc.  All rights reserved.
+ * Copyright (c) 2006-2009 Cisco Systems, Inc.  All rights reserved.
  * Copyright (c) 2006      Los Alamos National Security, LLC.  All rights
  *                         reserved. 
- * Copyright (c) 2008      Mellanox Technologies.  All rights reserved.
+ * Copyright (c) 2008-2009 Mellanox Technologies.  All rights reserved.
+ * Copyright (c) 2009      IBM Corporation.  All rights reserved.
  *
  * $COPYRIGHT$
  * 
@@ -24,10 +25,12 @@
 #include "ompi_config.h"
 
 #include "opal/dss/dss.h"
+#include "opal_stdint.h"
 #include "orte/util/show_help.h"
 #include "opal/util/error.h"
-#include "orte/mca/oob/base/base.h"
+#include "opal/util/output.h"
 #include "orte/mca/rml/rml.h"
+#include "orte/mca/rml/rml_types.h"
 #include "orte/mca/errmgr/errmgr.h"
 #include "orte/util/name_fns.h"
 #include "orte/runtime/orte_globals.h"
@@ -118,7 +121,7 @@ static int oob_component_query(mca_btl_openib_module_t *btl,
        transport_type member, then we must be < OFED v1.2, and
        therefore we must be IB. */   
 #if defined(HAVE_STRUCT_IBV_DEVICE_TRANSPORT_TYPE)
-    if (IBV_TRANSPORT_IB != btl->device->ib_dev->transport_type) {
+    if (BTL_OPENIB_CONNECT_BASE_CHECK_IF_NOT_IB(btl)) {
         opal_output_verbose(5, mca_btl_base_output,
                             "openib BTL: oob CPC only supported on InfiniBand; skipped on  %s:%d",
                             ibv_get_device_name(btl->device->ib_dev),
@@ -595,7 +598,7 @@ static int send_connect_data(mca_btl_base_endpoint_t* endpoint,
         ORTE_ERROR_LOG(rc);
         return rc;
     }
-    BTL_VERBOSE(("Sent QP Info, LID = %d, SUBNET = %016x\n",
+    BTL_VERBOSE(("Sent QP Info, LID = %d, SUBNET = %" PRIx64 "\n",
                  endpoint->endpoint_btl->lid, 
                  endpoint->subnet_id));
 
@@ -719,13 +722,16 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
         }
     }
     
-    BTL_VERBOSE(("Received QP Info,  LID = %d, SUBNET = %016x\n",
+    BTL_VERBOSE(("Received QP Info,  LID = %d, SUBNET = %" PRIx64 "\n",
                  rem_info.rem_lid, 
                  rem_info.rem_subnet_id));
     
     master = orte_util_compare_name_fields(ORTE_NS_CMP_ALL, ORTE_PROC_MY_NAME,
                                     process_name) > 0 ? true : false;
     
+    /* Need to protect the ib_procs list */
+    OPAL_THREAD_LOCK(&mca_btl_openib_component.ib_lock);
+
     for (ib_proc = (mca_btl_openib_proc_t*)
             opal_list_get_first(&mca_btl_openib_component.ib_procs);
         ib_proc != (mca_btl_openib_proc_t*)
@@ -778,6 +784,7 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
                just ignore this connection request */
             if (found && !master &&
                 MCA_BTL_IB_CLOSED != ib_endpoint->endpoint_state) {
+		OPAL_THREAD_UNLOCK(&mca_btl_openib_component.ib_lock);
                 return;
             }
         }
@@ -785,6 +792,7 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
         if (!found) {
             BTL_ERROR(("can't find suitable endpoint for this peer\n")); 
             mca_btl_openib_endpoint_invoke_error(NULL);
+	    OPAL_THREAD_UNLOCK(&mca_btl_openib_component.ib_lock);
             return; 
         }
         
@@ -808,6 +816,7 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
             if (OMPI_SUCCESS != rc) {
                 BTL_ERROR(("error in endpoint reply start connect"));
                 mca_btl_openib_endpoint_invoke_error(ib_endpoint);
+                OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                 break;
             }
             
@@ -816,6 +825,7 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
                RML events. Note: we increment it once peer active
                connection. */
             opal_progress_event_users_increment();
+            OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
             break;
              
         case MCA_BTL_IB_CONNECTING :
@@ -823,6 +833,7 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
             if (OMPI_SUCCESS != (rc = qp_connect_all(ib_endpoint))) {
                 BTL_ERROR(("endpoint connect error: %d", rc)); 
                 mca_btl_openib_endpoint_invoke_error(ib_endpoint);
+                OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                 break;
             }
            
@@ -831,32 +842,38 @@ static void rml_recv_cb(int status, orte_process_name_t* process_name,
 
                 /* Send him an ACK */
                 send_connect_data(ib_endpoint, ENDPOINT_CONNECT_RESPONSE);
+                OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
             } else {
                 send_connect_data(ib_endpoint, ENDPOINT_CONNECT_ACK);
                 /* Tell main BTL that we're done */
                 mca_btl_openib_endpoint_cpc_complete(ib_endpoint);
+                /* cpc complete unlock the endpoint */
              }
             break;
             
         case MCA_BTL_IB_WAITING_ACK:
             /* Tell main BTL that we're done */
             mca_btl_openib_endpoint_cpc_complete(ib_endpoint);
+            /* cpc complete unlock the endpoint */
             break;
             
         case MCA_BTL_IB_CONNECT_ACK:
             send_connect_data(ib_endpoint, ENDPOINT_CONNECT_ACK);
             /* Tell main BTL that we're done */
             mca_btl_openib_endpoint_cpc_complete(ib_endpoint);
+            /* cpc complete unlock the endpoint */
             break;
             
         case MCA_BTL_IB_CONNECTED:
+            OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
             break;
 
         default :
             BTL_ERROR(("Invalid endpoint state %d", endpoint_state));
             mca_btl_openib_endpoint_invoke_error(ib_endpoint);
+            OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
         }
-        OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
         break;
     }
+    OPAL_THREAD_UNLOCK(&mca_btl_openib_component.ib_lock);
 }
