@@ -1,6 +1,7 @@
 /*
  * Copyright (c) 2007-2009 Mellanox Technologies.  All rights reserved.
- * Copyright (c) 2008      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2009      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2009      IBM Corporation.  All rights reserved.
  *
  * $COPYRIGHT$
  *
@@ -11,11 +12,14 @@
 
 #include "ompi_config.h"
 
+#include "opal_stdint.h"
 #include "opal/dss/dss.h"
 #include "opal/util/error.h"
+#include "opal/util/output.h"
 #include "orte/util/show_help.h"
 #include "orte/util/name_fns.h"
 #include "orte/mca/rml/rml.h"
+#include "orte/mca/rml/rml_types.h"
 #include "orte/mca/errmgr/errmgr.h"
 #include "ompi/mca/dpm/dpm.h"
 
@@ -102,7 +106,7 @@ static int xoob_receive_connect_data(mca_btl_openib_rem_info_t *info, uint16_t *
         ORTE_ERROR_LOG(rc);
         return OMPI_ERROR;
     }
-    BTL_VERBOSE(("Recv unpack Message type  = %d", *message_type));
+    BTL_VERBOSE(("Recv unpack Message type  = %d\n", *message_type));
 
     BTL_VERBOSE(("unpacking %d of %d\n", cnt, OPAL_UINT64));
     rc = opal_dss.unpack(buffer, &info->rem_subnet_id, &cnt, OPAL_UINT64);
@@ -110,7 +114,7 @@ static int xoob_receive_connect_data(mca_btl_openib_rem_info_t *info, uint16_t *
         ORTE_ERROR_LOG(rc);
         return OMPI_ERROR;
     }
-    BTL_VERBOSE(("Recv unpack sid  = %d", info->rem_subnet_id));
+    BTL_VERBOSE(("Recv unpack sid  = %" PRIx64 "\n", info->rem_subnet_id));
 
     BTL_VERBOSE(("unpacking %d of %d\n", cnt, OPAL_UINT16));
     rc = opal_dss.unpack(buffer, &info->rem_lid, &cnt, OPAL_UINT16);
@@ -228,7 +232,7 @@ static int xoob_send_connect_data(mca_btl_base_endpoint_t* endpoint,
         return rc;
     }
 
-    BTL_VERBOSE(("Send pack sid = %d", endpoint->subnet_id));
+    BTL_VERBOSE(("Send pack sid = %" PRIx64 "\n", endpoint->subnet_id));
     BTL_VERBOSE(("packing %d of %d\n", 1, OPAL_UINT64));
     rc = opal_dss.pack(buffer, &endpoint->subnet_id, 1, OPAL_UINT64);
     if (ORTE_SUCCESS != rc) {
@@ -354,7 +358,7 @@ static int xoob_send_connect_data(mca_btl_base_endpoint_t* endpoint,
         return rc;
     }
 
-    BTL_VERBOSE(("Send QP Info, LID = %d, SUBNET = %016x\n, Message type = %d",
+    BTL_VERBOSE(("Send QP Info, LID = %d, SUBNET = %" PRIx64 ", Message type = %d",
                 endpoint->endpoint_btl->lid,
                 endpoint->subnet_id,
                 message_type));
@@ -586,7 +590,7 @@ static int xoob_recv_qp_create(mca_btl_openib_endpoint_t *endpoint, mca_btl_open
                     endpoint->xrc_recv_qp_num, strerror(ret), ret));
         return OMPI_ERROR;
     }
-#if OMPI_HAVE_THREADS
+#if OPAL_HAVE_THREADS
     if (APM_ENABLED) {
         mca_btl_openib_load_apm_xrc_rcv(endpoint->xrc_recv_qp_num, endpoint);
     }
@@ -652,9 +656,10 @@ static mca_btl_openib_endpoint_t* xoob_find_endpoint(orte_process_name_t* proces
     bool found = false;
 
     BTL_VERBOSE(("Searching for ep and proc with follow parameters:"
-                "jobid %d, vpid %d, sid %d, lid %d",
+                "jobid %d, vpid %d, sid %" PRIx64 ", lid %d",
                 process_name->jobid, process_name->vpid, subnet_id, lid));
     /* find ibproc */
+    OPAL_THREAD_LOCK(&mca_btl_openib_component.ib_lock);
     for (ib_proc = (mca_btl_openib_proc_t*)
             opal_list_get_first(&mca_btl_openib_component.ib_procs);
             ib_proc != (mca_btl_openib_proc_t*)
@@ -693,6 +698,7 @@ static mca_btl_openib_endpoint_t* xoob_find_endpoint(orte_process_name_t* proces
     } else {
             BTL_ERROR(("can't find suitable endpoint for this peer\n"));
     }
+    OPAL_THREAD_UNLOCK(&mca_btl_openib_component.ib_lock);
     return ib_endpoint;
 }
 
@@ -708,10 +714,10 @@ static void xoob_restart_connect(mca_btl_base_endpoint_t *endpoint)
         case MCA_BTL_IB_ADDR_CONNECTED:
             /* so we have the send qp, we just need the recive site.
              * Send request for SRQ numbers */
-            BTL_VERBOSE(("Restart The IB addr: sid %d lid %d"
-                        "in MCA_BTL_IB_ADDR_CONNECTED status,"
-                        " Changing to MCA_BTL_IB_ADDR_CLOSED and starting from scratch\n",
-                        endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
+            BTL_VERBOSE(("Restart The IB addr: sid %" PRIx64 " lid %d"
+                         "in MCA_BTL_IB_ADDR_CONNECTED status,"
+                         " Changing to MCA_BTL_IB_ADDR_CLOSED and starting from scratch\n",
+                         endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
             /* Switching back to closed and starting from scratch */
             endpoint->ib_addr->status = MCA_BTL_IB_ADDR_CLOSED;
             /* destroy the qp */
@@ -722,10 +728,10 @@ static void xoob_restart_connect(mca_btl_base_endpoint_t *endpoint)
                 BTL_ERROR(("Failed to destroy QP"));
         case MCA_BTL_IB_ADDR_CLOSED:
         case MCA_BTL_IB_ADDR_CONNECTING:
-            BTL_VERBOSE(("Restart The IB addr: sid %d lid %d"
-                        "in MCA_BTL_IB_ADDR_CONNECTING or MCA_BTL_IB_ADDR_CLOSED status,"
-                        " starting from scratch\n",
-                        endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
+            BTL_VERBOSE(("Restart The IB addr: sid %" PRIx64 " lid %d"
+                         "in MCA_BTL_IB_ADDR_CONNECTING or MCA_BTL_IB_ADDR_CLOSED status,"
+                         " starting from scratch\n",
+                         endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
             OPAL_THREAD_UNLOCK(&endpoint->ib_addr->addr_lock);
             /* xoob_module_start_connect() should automaticly handle all other cases */
             if (OMPI_SUCCESS != xoob_module_start_connect(NULL, endpoint))
@@ -794,7 +800,7 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
     /* Processing message */
     switch (message_type) {
         case ENDPOINT_XOOB_CONNECT_REQUEST:
-            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_REQUEST: lid %d, sid %d, rlid %d\n",
+            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_REQUEST: lid %d, sid %" PRIx64 ", rlid %d\n",
                         rem_info.rem_lid,
                         rem_info.rem_subnet_id,
                         requested_lid));
@@ -802,8 +808,9 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
                     requested_lid, message_type);
             if ( NULL == ib_endpoint) {
                 BTL_ERROR(("Got ENDPOINT_XOOB_CONNECT_REQUEST."
-                            " Failed to find endpoint with subnet %d and LID %d",
-                            rem_info.rem_subnet_id,requested_lid));
+                           " Failed to find endpoint with subnet %" PRIx64 
+                           " and LID %d",
+                           rem_info.rem_subnet_id,requested_lid));
                 mca_btl_openib_endpoint_invoke_error(NULL);
                 return;
             }
@@ -812,6 +819,7 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
             if (OMPI_SUCCESS != mca_btl_openib_endpoint_post_recvs(ib_endpoint)) {
                 BTL_ERROR(("Failed to post on XRC SRQs"));
                 mca_btl_openib_endpoint_invoke_error(NULL);
+                OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                 return;
             }
             /* we should create qp and send the info + srq to requestor */
@@ -819,6 +827,7 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
             if (OMPI_SUCCESS != rc) {
                 BTL_ERROR(("error in endpoint reply start connect"));
                 mca_btl_openib_endpoint_invoke_error(NULL);
+                OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                 return;
             }
             /* enable pooling for this btl */
@@ -826,14 +835,14 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
             break;
         case ENDPOINT_XOOB_CONNECT_XRC_REQUEST:
             /* pasha we don't need the remote lid here ??*/
-            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_XRC_REQUEST: lid %d, sid %d\n",
+            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_XRC_REQUEST: lid %d, sid %" PRIx64 "\n",
                         rem_info.rem_lid,
                         rem_info.rem_subnet_id));
             ib_endpoint = xoob_find_endpoint(process_name,rem_info.rem_subnet_id,
                     requested_lid, message_type);
             if ( NULL == ib_endpoint) {
                 BTL_ERROR(("Got ENDPOINT_XOOB_CONNECT_XRC_REQUEST."
-                            " Failed to find endpoint with subnet %d and LID %d",
+                            " Failed to find endpoint with subnet %" PRIx64 " and LID %d",
                             rem_info.rem_subnet_id,requested_lid));
                 mca_btl_openib_endpoint_invoke_error(NULL);
                 return;
@@ -849,6 +858,7 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
                 if (OMPI_SUCCESS != rc) {
                     BTL_ERROR(("error in endpoint reply start connect"));
                     mca_btl_openib_endpoint_invoke_error(ib_endpoint);
+                    OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                     return;
                 }
                 OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
@@ -859,6 +869,7 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
                 if (OMPI_SUCCESS != rc) {
                     BTL_ERROR(("error in endpoint reply start connect"));
                     mca_btl_openib_endpoint_invoke_error(ib_endpoint);
+                    OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                     return;
                 }
                 OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
@@ -866,14 +877,14 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
             /* enable pooling for this btl */
             break;
         case ENDPOINT_XOOB_CONNECT_RESPONSE:
-            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_RESPONSE: lid %d, sid %d\n",
+            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_RESPONSE: lid %d, sid %" PRIx64 "\n",
                         rem_info.rem_lid,
                         rem_info.rem_subnet_id));
             ib_endpoint = xoob_find_endpoint(process_name, rem_info.rem_subnet_id,
                     rem_info.rem_lid, message_type);
             if ( NULL == ib_endpoint) {
                 BTL_ERROR(("Got ENDPOINT_XOOB_CONNECT_RESPONSE."
-                            " Failed to find endpoint with subnet %d and LID %d",
+                            " Failed to find endpoint with subnet %" PRIx64 " and LID %d",
                             rem_info.rem_subnet_id,rem_info.rem_lid));
                 mca_btl_openib_endpoint_invoke_error(NULL);
                 return;
@@ -884,27 +895,30 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
             /* update ib_addr with remote qp number */
             ib_endpoint->ib_addr->remote_xrc_rcv_qp_num =
                 ib_endpoint->rem_info.rem_qps->rem_qp_num;
-            BTL_VERBOSE(("rem_info: lid %d, sid %d ep %d %d",
-                        rem_info.rem_lid,
-                        rem_info.rem_subnet_id,
-                        ib_endpoint->rem_info.rem_lid,ib_endpoint->rem_info.rem_subnet_id));
+            BTL_VERBOSE(("rem_info: lid %d, sid %" PRIx64 
+                         " ep %d %" PRIx64 "\n",
+                         rem_info.rem_lid,
+                         rem_info.rem_subnet_id,
+                         ib_endpoint->rem_info.rem_lid,
+                         ib_endpoint->rem_info.rem_subnet_id));
             if (OMPI_SUCCESS != xoob_send_qp_connect(ib_endpoint, &rem_info)) {
                 BTL_ERROR(("Failed to connect  endpoint\n"));
                 mca_btl_openib_endpoint_invoke_error(NULL);
+                OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
                 return;
             }
             mca_btl_openib_endpoint_cpc_complete(ib_endpoint);
-            OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
+            /* cpc complete unlock the endpoint */
             break;
         case ENDPOINT_XOOB_CONNECT_XRC_RESPONSE:
-            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_XRC_RESPONSE: lid %d, sid %d\n",
+            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_XRC_RESPONSE: lid %d, sid %" PRIx64 "\n",
                         rem_info.rem_lid,
                         rem_info.rem_subnet_id));
             ib_endpoint = xoob_find_endpoint(process_name, rem_info.rem_subnet_id,
                     rem_info.rem_lid, message_type);
             if ( NULL == ib_endpoint) {
                 BTL_ERROR(("Got ENDPOINT_XOOB_CONNECT_XRC_RESPONSE."
-                            " Failed to find endpoint with subnet %d and LID %d",
+                            " Failed to find endpoint with subnet %" PRIx64 " and LID %d",
                             rem_info.rem_subnet_id,rem_info.rem_lid));
                 mca_btl_openib_endpoint_invoke_error(NULL);
                 return;
@@ -913,19 +927,19 @@ static void xoob_rml_recv_cb(int status, orte_process_name_t* process_name,
             /* we got srq numbers on our request */
             XOOB_SET_REMOTE_INFO(ib_endpoint->rem_info, rem_info);
             mca_btl_openib_endpoint_cpc_complete(ib_endpoint);
-            OPAL_THREAD_UNLOCK(&ib_endpoint->endpoint_lock);
+            /* cpc complete unlock the endpoint */
             break;
         case ENDPOINT_XOOB_CONNECT_XRC_NR_RESPONSE:
             /* The XRC recv site already was destroyed so we need
              * start to bringup the connection from scratch  */
-            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_XRC_NR_RESPONSE: lid %d, sid %d\n",
+            BTL_VERBOSE(("Received ENDPOINT_XOOB_CONNECT_XRC_NR_RESPONSE: lid %d, sid %" PRIx64 "\n",
                         rem_info.rem_lid,
                         rem_info.rem_subnet_id));
             ib_endpoint = xoob_find_endpoint(process_name, rem_info.rem_subnet_id,
                     rem_info.rem_lid, message_type);
             if ( NULL == ib_endpoint) {
                 BTL_ERROR(("Got ENDPOINT_XOOB_CONNECT_XRC_NR_RESPONSE."
-                            " Failed to find endpoint with subnet %d and LID %d",
+                            " Failed to find endpoint with subnet %" PRIx64 " and LID %d",
                             rem_info.rem_subnet_id,rem_info.rem_lid));
                 mca_btl_openib_endpoint_invoke_error(NULL);
                 return;
@@ -1029,7 +1043,7 @@ static int xoob_module_start_connect(ompi_btl_openib_connect_base_module_t *cpc,
     OPAL_THREAD_LOCK(&endpoint->ib_addr->addr_lock);
     switch (endpoint->ib_addr->status) {
         case MCA_BTL_IB_ADDR_CLOSED:
-            BTL_VERBOSE(("The IB addr: sid %d lid %d"
+            BTL_VERBOSE(("The IB addr: sid %" PRIx64 " lid %d"
                         "in MCA_BTL_IB_ADDR_CLOSED status,"
                         " sending ENDPOINT_XOOB_CONNECT_REQUEST\n",
                         endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
@@ -1046,7 +1060,7 @@ static int xoob_module_start_connect(ompi_btl_openib_connect_base_module_t *cpc,
             }
             break;
         case MCA_BTL_IB_ADDR_CONNECTING:
-            BTL_VERBOSE(("The IB addr: sid %d lid %d"
+            BTL_VERBOSE(("The IB addr: sid %" PRIx64 " lid %d"
                         "in MCA_BTL_IB_ADDR_CONNECTING status,"
                         " Subscribing to this address\n",
                         endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
@@ -1057,7 +1071,7 @@ static int xoob_module_start_connect(ompi_btl_openib_connect_base_module_t *cpc,
         case MCA_BTL_IB_ADDR_CONNECTED:
             /* so we have the send qp, we just need the recive site.
              * Send request for SRQ numbers */
-            BTL_VERBOSE(("The IB addr: sid %d lid %d"
+            BTL_VERBOSE(("The IB addr: sid %" PRIx64 " lid %d"
                         "in MCA_BTL_IB_ADDR_CONNECTED status,"
                         " sending ENDPOINT_XOOB_CONNECT_XRC_REQUEST\n",
                         endpoint->ib_addr->subnet_id,endpoint->ib_addr->lid));
