@@ -13,6 +13,7 @@
  * Copyright (c) 2007      Los Alamos National Security, LLC.  All rights
  *                         reserved. 
  * Copyright (c) 2008      Sun Microsystems, Inc.  All rights reserved.
+ * Copyright (c) 2009      University of Houston.  All rights reserved.
  * $COPYRIGHT$
  * 
  * Additional copyrights may follow
@@ -31,6 +32,7 @@
 
 #include "ompi_config.h"
 
+#include "opal/class/opal_list.h"
 #include "opal/class/opal_hash_table.h"
 
 BEGIN_C_DECLS
@@ -56,9 +58,27 @@ OMPI_DECLSPEC extern int ompi_mpi_thread_provided;
 /** Identifier of the main thread */
 OMPI_DECLSPEC extern struct opal_thread_t *ompi_mpi_main_thread;
 
-/** Did we setup maffinity in MPI_INIT (and therefore need to shut
-    it down during MPI_FINALIZE)? */
-OMPI_DECLSPEC extern bool ompi_mpi_maffinity_setup;
+
+/** Bitflags to be used for the modex exchange for the various thread
+ *  levels. Required to support heterogeneous environments */
+#define OMPI_THREADLEVEL_SINGLE_BF     0x00000001
+#define OMPI_THREADLEVEL_FUNNELED_BF   0x00000002
+#define OMPI_THREADLEVEL_SERIALIZED_BF 0x00000004
+#define OMPI_THREADLEVEL_MULTIPLE_BF   0x00000008
+
+#define OMPI_THREADLEVEL_SET_BITFLAG(threadlevelin,threadlevelout) { \
+    if ( MPI_THREAD_SINGLE == threadlevelin ) {                 \
+        threadlevelout |= OMPI_THREADLEVEL_SINGLE_BF;           \
+    } else if ( MPI_THREAD_FUNNELED == threadlevelin ) {        \
+        threadlevelout |= OMPI_THREADLEVEL_FUNNELED_BF;         \
+    } else if ( MPI_THREAD_SERIALIZED == threadlevelin ) {      \
+        threadlevelout |= OMPI_THREADLEVEL_SERIALIZED_BF;       \
+    } else if ( MPI_THREAD_MULTIPLE == threadlevelin ) {       \
+        threadlevelout |= OMPI_THREADLEVEL_MULTIPLE_BF;         \
+    }}
+
+
+#define OMPI_THREADLEVEL_IS_MULTIPLE(threadlevel) (threadlevel & OMPI_THREADLEVEL_MULTIPLE_BF)
 
 /** Do we want to be warned on fork or not? */
 OMPI_DECLSPEC extern bool ompi_warn_on_fork;
@@ -116,7 +136,7 @@ int ompi_mpi_finalize(void);
 /**
  * Abort the processes of comm
  */
-int ompi_mpi_abort(struct ompi_communicator_t* comm,
+OMPI_DECLSPEC int ompi_mpi_abort(struct ompi_communicator_t* comm,
                    int errcode, bool kill_remote_of_intercomm);
 
 /**
