@@ -11,7 +11,8 @@
  *                         All rights reserved.
  * Copyright (c) 2007-2009 Sun Microsystems, Inc.  All rights reserved.
  * Copyright (c) 2007      Evergrid, Inc. All rights reserved.
- * Copyright (c) 2008      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2008-2010 Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2010      IBM Corporation.  All rights reserved.
  *
  * $COPYRIGHT$
  *
@@ -23,7 +24,11 @@
 
 #include "orte_config.h"
 #include "orte/constants.h"
+#include "orte/types.h"
 
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
 #include <stdlib.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -66,14 +71,16 @@
 
 #include "opal/mca/maffinity/base/base.h"
 #include "opal/mca/paffinity/base/base.h"
-#include "opal/class/opal_value_array.h"
+#include "opal/class/opal_pointer_array.h"
+#include "opal/util/opal_environ.h"
 
 #include "orte/util/show_help.h"
 #include "orte/runtime/orte_wait.h"
 #include "orte/runtime/orte_globals.h"
 #include "orte/mca/errmgr/errmgr.h"
-#include "orte/mca/ess/base/base.h"
+#include "orte/mca/ess/ess.h"
 #include "orte/mca/iof/base/iof_base_setup.h"
+#include "orte/mca/plm/plm.h"
 #include "orte/util/name_fns.h"
 
 #include "orte/mca/odls/base/odls_private.h"
@@ -83,7 +90,7 @@
  * External Interface
  */
 static int orte_odls_default_launch_local_procs(opal_buffer_t *data);
-static int orte_odls_default_kill_local_procs(orte_jobid_t job, bool set_state);
+static int orte_odls_default_kill_local_procs(opal_pointer_array_t *procs, bool set_state);
 static int orte_odls_default_signal_local_procs(const orte_process_name_t *proc, int32_t signal);
 
 static void set_handler_default(int sig);
@@ -94,8 +101,7 @@ orte_odls_base_module_t orte_odls_default_module = {
     orte_odls_default_kill_local_procs,
     orte_odls_default_signal_local_procs,
     orte_odls_base_default_deliver_message,
-    orte_odls_base_default_require_sync,
-    orte_odls_base_default_collect_data
+    orte_odls_base_default_require_sync
 };
 
 /* convenience macro for erroring out */
@@ -110,11 +116,11 @@ orte_odls_base_module_t orte_odls_default_module = {
 #define ORTE_ODLS_IF_BIND_NOT_REQD(n)                               \
     do {                                                            \
         if (ORTE_BINDING_NOT_REQUIRED(jobdat->policy)) {            \
-            if (orte_odls_globals.report_bindings) {                \
+            if (orte_report_bindings) {                             \
                 orte_show_help("help-odls-default.txt",             \
                                "odls-default:binding-not-avail",    \
                                true, orte_process_info.nodename,    \
-                               (n), context->app);		    \
+                               (n), context->app);                  \
             }                                                       \
             goto LAUNCH_PROCS;                                      \
         }                                                           \
@@ -133,6 +139,9 @@ static bool odls_default_child_died(pid_t pid, unsigned int timeout, int *exit_s
     do {
         ret = waitpid(pid, exit_status, WNOHANG);
         if (pid == ret) {
+            OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
+                                 "%s odls:default:WAITPID INDICATES PROC %d IS DEAD",
+                                 ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), (int)pid));
             /* It died -- return success */
             return true;
         } else if (0 == ret) {
@@ -144,10 +153,16 @@ static bool odls_default_child_died(pid_t pid, unsigned int timeout, int *exit_s
              * by waitpid in this case, so we cannot check it - just assume
              * the proc has indeed died
              */
+            OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
+                                 "%s odls:default:WAITPID INDICATES PROC %d HAS ALREADY EXITED",
+                                 ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), (int)pid));
             return true;
         } else if (-1 == ret && ECHILD == errno) {
             /* The pid no longer exists, so we'll call this "good
                enough for government work" */
+            OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
+                                 "%s odls:default:WAITPID INDICATES PID %d NO LONGER EXISTS",
+                                 ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), (int)pid));
             return true;
         }
         
@@ -170,17 +185,28 @@ static bool odls_default_child_died(pid_t pid, unsigned int timeout, int *exit_s
 
 static int odls_default_kill_local(pid_t pid, int signum)
 {
-    if (0 != kill(pid, signum)) {
-        if (ESRCH != errno) return errno;
+    if (orte_forward_job_control) {
+        pid = -pid;
     }
+    if (0 != kill(pid, signum)) {
+        if (ESRCH != errno) {
+            OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
+                                 "%s odls:default:SENT KILL %d TO PID %d GOT ERRNO %d",
+                                 ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), signum, (int)pid, errno));
+            return errno;
+        }
+    }
+    OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
+                         "%s odls:default:SENT KILL %d TO PID %d SUCCESS",
+                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), signum, (int)pid));
     return 0;
 }
 
-int orte_odls_default_kill_local_procs(orte_jobid_t job, bool set_state)
+int orte_odls_default_kill_local_procs(opal_pointer_array_t *procs, bool set_state)
 {
     int rc;
     
-    if (ORTE_SUCCESS != (rc = orte_odls_base_default_kill_local_procs(job, set_state,
+    if (ORTE_SUCCESS != (rc = orte_odls_base_default_kill_local_procs(procs, set_state,
                                     odls_default_kill_local, odls_default_child_died))) {
         ORTE_ERROR_LOG(rc);
         return rc;
@@ -210,12 +236,12 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
     int target_socket, npersocket, logical_skt;
     int logical_cpu, phys_core, phys_cpu, ncpu;
     bool bound = false;
-    orte_pmap_t *pmap;
+    char *param, *tmp;
     
     if (NULL != child) {
         /* should pull this information from MPIRUN instead of going with
          default */
-        opts.usepty = OMPI_ENABLE_PTY_SUPPORT;
+        opts.usepty = OPAL_ENABLE_PTY_SUPPORT;
         
         /* do we want to setup stdin? */
         if (NULL != child &&
@@ -245,8 +271,10 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
      the pipe, then the child was letting us know that it failed. */
     if (pipe(p) < 0) {
         ORTE_ERROR_LOG(ORTE_ERR_SYS_LIMITS_PIPES);
-        child->state = ORTE_PROC_STATE_FAILED_TO_START;
-        child->exit_code = ORTE_ERR_SYS_LIMITS_PIPES;
+        if (NULL != child) {
+            child->state = ORTE_PROC_STATE_FAILED_TO_START;
+            child->exit_code = ORTE_ERR_SYS_LIMITS_PIPES;
+        }
         return ORTE_ERR_SYS_LIMITS_PIPES;
     }
     
@@ -258,18 +286,27 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
     
     if(pid < 0) {
         ORTE_ERROR_LOG(ORTE_ERR_SYS_LIMITS_CHILDREN);
-        child->state = ORTE_PROC_STATE_FAILED_TO_START;
-        child->exit_code = ORTE_ERR_SYS_LIMITS_CHILDREN;
+        if (NULL != child) {
+            child->state = ORTE_PROC_STATE_FAILED_TO_START;
+            child->exit_code = ORTE_ERR_SYS_LIMITS_CHILDREN;
+        }
         return ORTE_ERR_SYS_LIMITS_CHILDREN;
     }
     
-    if (child->pid == 0) {
+    if (pid == 0) {
         long fd, fdmax = sysconf(_SC_OPEN_MAX);
+
+        if (orte_forward_job_control) {
+            /* Set a new process group for this child, so that a
+               SIGSTOP can be sent to it without being sent to the
+               orted. */
+            setpgid(0, 0);
+        }
         
         /* Setup the pipe to be close-on-exec */
         close(p[0]);
         fcntl(p[1], F_SETFD, FD_CLOEXEC);
-
+        
         if (NULL != child) {
             /*  setup stdout/stderr so that any error messages that we may
              print out will get displayed back at orterun.
@@ -286,18 +323,6 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                 ORTE_ODLS_ERROR_OUT(i);
             }
             
-            /* get the pmap object for this child */
-            if (NULL == (pmap = (orte_pmap_t*)opal_value_array_get_item(&jobdat->procmap, child->name->vpid))) {
-                ORTE_ERROR_LOG(ORTE_ERR_NOT_FOUND);
-                ORTE_ODLS_ERROR_OUT(ORTE_ERR_NOT_FOUND);
-            }
-            /* get the local rank */
-            if (ORTE_LOCAL_RANK_INVALID == (lrank = pmap->local_rank)) {
-                orte_show_help("help-odls-default.txt",
-                               "odls-default:invalid-local-rank", true);
-                ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
-            }
-            
             /* Setup process affinity.  First check to see if a slot list was
              * specified.  If so, use it.  If no slot list was specified,
              * that's not an error -- just fall through and try the next
@@ -314,7 +339,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                                    "odls-default:multiple-paffinity-schemes", true, child->slot_list);
                     ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
                 }
-                if (orte_odls_globals.report_bindings) {
+                if (orte_report_bindings) {
                     opal_output(0, "%s odls:default:fork binding child %s to slot_list %s",
                                 ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                                 ORTE_NAME_PRINT(child->name), child->slot_list);
@@ -342,13 +367,18 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                                      ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                                      ORTE_NAME_PRINT(child->name),
                                      (int)jobdat->cpus_per_rank, (int)jobdat->stride));
-
                 /* get the node rank */
-                if (ORTE_NODE_RANK_INVALID == (nrank = pmap->node_rank)) {
+                if (ORTE_NODE_RANK_INVALID == (nrank = orte_ess.get_node_rank(child->name))) {
                     orte_show_help("help-odls-default.txt",
                                    "odls-default:invalid-node-rank", true);
                     ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
                 }
+                /* get the local rank */
+                if (ORTE_LOCAL_RANK_INVALID == (lrank = orte_ess.get_local_rank(child->name))) {
+                    orte_show_help("help-odls-default.txt",
+                                   "odls-default:invalid-local-rank", true);
+                    ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
+                }
                 /* init the mask */
                 OPAL_PAFFINITY_CPU_ZERO(mask);
                 if (ORTE_MAPPING_NPERXXX & jobdat->policy) {
@@ -367,7 +397,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                     }
                     if (orte_odls_globals.bound) {
                         /* if we are bound, use this as an index into our available sockets */
-                        for (target_socket=0, n=0; target_socket < opal_bitmap_size(&orte_odls_globals.sockets) && n < logical_skt; target_socket++) {
+                        for (n=target_socket=0; target_socket < opal_bitmap_size(&orte_odls_globals.sockets) && n < logical_skt; target_socket++) {
                             if (opal_bitmap_is_set_bit(&orte_odls_globals.sockets, target_socket)) {
                                 n++;
                             }
@@ -464,7 +494,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                         /* increment logical cpu */
                         logical_cpu += jobdat->stride;
                     }
-                    if (orte_odls_globals.report_bindings) {
+                    if (orte_report_bindings) {
                         opal_output(0, "%s odls:default:fork binding child %s to socket %d cpus %04lx",
                                     ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                                     ORTE_NAME_PRINT(child->name), target_socket, mask.bitmask[0]);
@@ -479,14 +509,14 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                              * against our available cores
                              */
                             ncpu = 0;
-                            for (i=0; i < orte_odls_globals.num_processors && ncpu <= logical_cpu; i++) {
+                            for (i=0; i < OPAL_PAFFINITY_BITMASK_CPU_MAX && ncpu <= logical_cpu; i++) {
                                 if (OPAL_PAFFINITY_CPU_ISSET(i, orte_odls_globals.my_cores)) {
                                     ncpu++;
                                     phys_cpu = i;
                                 }
                             }
                             /* if we don't have enough processors, that is an error */
-                            if (ncpu < logical_cpu) {
+                            if (ncpu <= logical_cpu) {
                                 ORTE_ODLS_IF_BIND_NOT_REQD("bind-to-core");
                                 orte_show_help("help-odls-default.txt",
                                                "odls-default:not-enough-resources", true,
@@ -500,7 +530,15 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                              * physical cpu
                              */
                             phys_cpu = opal_paffinity_base_get_physical_processor_id(logical_cpu);
-                            if (0 > phys_cpu) {
+                            if (OPAL_ERROR == phys_cpu){
+                                /* No processor to bind to so error out */
+                                ORTE_ODLS_IF_BIND_NOT_REQD("bind-to-core");
+                                orte_show_help("help-odls-default.txt",
+                                               "odls-default:not-enough-resources", true,
+                                               "processors", orte_process_info.nodename,
+                                               "bind-to-core", context->app);
+                                ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
+                            } else if (0 > phys_cpu) {
                                 ORTE_ODLS_IF_BIND_NOT_REQD("bind-to-core");
                                 orte_show_help("help-odls-default.txt",
                                                "odls-default:invalid-phys-cpu", true);
@@ -511,7 +549,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                         /* increment logical cpu */
                         logical_cpu += jobdat->stride;
                     }
-                    if (orte_odls_globals.report_bindings) {
+                    if (orte_report_bindings) {
                         opal_output(0, "%s odls:default:fork binding child %s to cpus %04lx",
                                     ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                                     ORTE_NAME_PRINT(child->name), mask.bitmask[0]);
@@ -533,6 +571,11 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                 /* layout this process across the sockets based on
                  * the provided mapping policy
                  */
+                if (ORTE_LOCAL_RANK_INVALID == (lrank = orte_ess.get_local_rank(child->name))) {
+                    orte_show_help("help-odls-default.txt",
+                                   "odls-default:invalid-local-rank", true);
+                    ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
+                }
                 if (ORTE_MAPPING_NPERXXX & jobdat->policy) {
                     /* we need to balance the children from this job across the available sockets */
                     npersocket = jobdat->num_local_procs / orte_odls_globals.num_sockets;
@@ -549,7 +592,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                     }
                     if (orte_odls_globals.bound) {
                         /* if we are bound, use this as an index into our available sockets */
-                        for (target_socket=0; target_socket < opal_bitmap_size(&orte_odls_globals.sockets) && n < logical_skt; target_socket++) {
+                        for (target_socket=0, n = 0; target_socket < opal_bitmap_size(&orte_odls_globals.sockets) && n < logical_skt; target_socket++) {
                             if (opal_bitmap_is_set_bit(&orte_odls_globals.sockets, target_socket)) {
                                 n++;
                             }
@@ -718,7 +761,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                                    "odls-default:could-not-bind-to-socket", true);
                     ORTE_ODLS_ERROR_OUT(ORTE_ERR_FATAL);
                 }
-                if (orte_odls_globals.report_bindings) {
+                if (orte_report_bindings) {
                     opal_output(0, "%s odls:default:fork binding child %s to socket %d cpus %04lx",
                                 ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                                 ORTE_NAME_PRINT(child->name), target_socket, mask.bitmask[0]);
@@ -759,10 +802,25 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
             close(fdnull);
         }
         
-    LAUNCH_PROCS:
+LAUNCH_PROCS:
+        /* if we are bound, report it */
+        if (opal_paffinity_base_bound) {
+            param = mca_base_param_environ_variable("paffinity","base","bound");
+            opal_setenv(param, "1", true, &environ_copy);
+            free (param);
+            /* and provide a char representation of what we did */
+            tmp = opal_paffinity_base_print_binding(mask);
+            if (NULL != tmp) {
+                param = mca_base_param_environ_variable("paffinity","base","applied_binding");
+                opal_setenv(param, tmp, true, &environ_copy);
+                free(tmp);
+            }
+        }
+
         /* close all file descriptors w/ exception of
-         stdin/stdout/stderr and the pipe used for the IOF INTERNAL
-         messages */
+         * stdin/stdout/stderr and the pipe used for the IOF INTERNAL
+         * messages
+         */
         for(fd=3; fd<fdmax; fd++) {
             if (fd != opts.p_internal[1]) {
                 close(fd);
@@ -802,7 +860,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                        true, context->app, strerror(errno));
         exit(1);
     } else {
-        
+
         if (NULL != child && (ORTE_JOB_CONTROL_FORWARD_OUTPUT & jobdat->controls)) {
             /* connect endpoints IOF */
             rc = orte_iof_base_setup_parent(child->name, &opts);
@@ -811,7 +869,7 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                 return rc;
             }
         }
-        
+
         /* Wait to read something from the pipe or close */
         close(p[1]);
         while (1) {
@@ -821,9 +879,12 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                 if (errno == EINTR) {
                     continue;
                 }
+                
                 /* Other errno's are bad */
-                child->state = ORTE_PROC_STATE_FAILED_TO_START;
-                child->exit_code = ORTE_ERR_PIPE_READ_FAILURE;
+                if (NULL != child) {
+                    child->state = ORTE_PROC_STATE_FAILED_TO_START;
+                    child->exit_code = ORTE_ERR_PIPE_READ_FAILURE;
+                }
                 
                 OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
                                      "%s odls:default:fork got code %d back from child",
@@ -835,15 +896,17 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                 break;
             } else {
                 /*  Doh -- child failed.
-                 Let the calling function
-                 know about the failure.  The actual exit status of child proc
-                 cannot be found here - all we can do is report the ORTE error
-                 code that was reported back to us. The calling func needs to report the
-                 failure to launch this process through the SMR or else
-                 everyone else will hang.
-                 */
-                child->state = ORTE_PROC_STATE_FAILED_TO_START;
-                child->exit_code = i;
+                    Let the calling function
+                    know about the failure.  The actual exit status of child proc
+                    cannot be found here - all we can do is report the ORTE error
+                    code that was reported back to us. The calling func needs to report the
+                    failure to launch this process through the SMR or else
+                    everyone else will hang.
+                */
+                if (NULL != child) {
+                    child->state = ORTE_PROC_STATE_FAILED_TO_START;
+                    child->exit_code = i;
+                }
                 
                 OPAL_OUTPUT_VERBOSE((2, orte_odls_globals.output,
                                      "%s odls:default:fork got code %d back from child",
@@ -852,10 +915,12 @@ static int odls_default_fork_local_proc(orte_app_context_t* context,
                 return ORTE_ERR_FAILED_TO_START;
             }
         }
-        
-        /* set the proc state to LAUNCHED */
-        child->state = ORTE_PROC_STATE_LAUNCHED;
-        child->alive = true;
+
+        if (NULL != child) {
+            /* set the proc state to LAUNCHED */
+            child->state = ORTE_PROC_STATE_LAUNCHED;
+            child->alive = true;
+        }
         close(p[0]);
     }
     
@@ -871,6 +936,7 @@ int orte_odls_default_launch_local_procs(opal_buffer_t *data)
 {
     int rc;
     orte_jobid_t job;
+    orte_job_t *jdata;
 
     /* construct the list of children we are to launch */
     if (ORTE_SUCCESS != (rc = orte_odls_base_default_construct_child_list(data, &job))) {
@@ -888,6 +954,23 @@ int orte_odls_default_launch_local_procs(opal_buffer_t *data)
         goto CLEANUP;
     }
     
+    /* look up job data object */
+    if (NULL != (jdata = orte_get_job_data_object(job))) {
+        if (jdata->state & ORTE_JOB_STATE_SUSPENDED) {
+            if (ORTE_PROC_IS_HNP) {
+                /* Have the plm send the signal to all the nodes.
+                   If the signal arrived before the orteds started,
+                   then they won't know to suspend their procs.
+                   The plm also arranges for any local procs to
+                   be signaled.
+                 */
+                orte_plm.signal_job(jdata->jobid, SIGTSTP);
+            } else {
+                orte_odls_default_signal_local_procs(NULL, SIGTSTP);
+            }
+        }
+    }
+
 CLEANUP:
    
     return rc;
@@ -917,7 +1000,12 @@ static int send_signal(pid_t pid, int signal)
                          "%s sending signal %d to pid %ld",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                          signal, (long)pid));
-    
+
+    if (orte_forward_job_control) {
+        /* Send the signal to the process group rather than the
+           process.  The child is the leader of its process group. */
+        pid = -pid;
+    }
     if (kill(pid, signal) != 0) {
         switch(errno) {
             case EINVAL:
