@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
+ * Copyright (c) 2004-2009 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
  * Copyright (c) 2004-2005 The University of Tennessee and The University
@@ -10,6 +10,7 @@
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
  * Copyright (c) 2006      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2010      IBM Corporation.  All rights reserved.
  * $COPYRIGHT$
  * 
  * Additional copyrights may follow
@@ -18,22 +19,15 @@
  */
 /* Much of the code in this file is taken from the file ll_get_machine_list.c, 
  * which is provided by IBM as part of their sample programs for LoadLeveler
- * in the samples/llmpich directory. The documentation has the following license:
- * COPYRIGHT LICENSE:
- * This information contains sample application programs in source language, which 
- * illustrate programming techniques on various operating platforms. You may copy,
- * modify, and distribute these sample programs in any form without payment to 
- * IBM, for the purposes of developing, using, marketing or distributing 
- * application programs conforming to the application programming interface for 
- * the operating platform for which the sample programs are written. These 
- * examples have not been thoroughly tested under all conditions. IBM, 
- * therefore, cannot guarantee or imply reliability, serviceability, or
- * function of these programs.
+ * in the samples/llmpich directory.
+ * 
+ * IBM has approved the release of the sample code for Loadleveler under the 
+ * BSD license.  Consequently, a more restrictive licensing clause that was
+ * originally associated with the sample code and replicated here has been
+ * removed.
  */
  
 #include "orte_config.h"
-#include "orte/orte_constants.h"
-#include "orte/orte_types.h"
 
 #include <errno.h>
 #include <unistd.h>
@@ -44,10 +38,9 @@
 #include "opal/util/argv.h"
 #include "opal/util/output.h"
 
-#include "orte/dss/dss.h"
-#include "orte/mca/gpr/gpr.h"
-#include "orte/mca/rmgr/rmgr.h"
 #include "orte/mca/errmgr/errmgr.h"
+#include "orte/runtime/orte_globals.h"
+#include "orte/constants.h"
 
 #include "orte/mca/ras/base/ras_private.h"
 #include "ras_loadleveler.h"
@@ -56,8 +49,7 @@
 /*
  * Local functions
  */
-static int orte_ras_loadleveler_allocate(orte_jobid_t jobid, opal_list_t *attributes);
-static int orte_ras_loadleveler_deallocate(orte_jobid_t jobid);
+static int orte_ras_loadleveler_allocate(opal_list_t *nodes);
 static int orte_ras_loadleveler_finalize(void);
 static int orte_ras_loadleveler_get_hostlist(int * num_hosts, char*** hostlist);
 
@@ -67,11 +59,6 @@ static int orte_ras_loadleveler_get_hostlist(int * num_hosts, char*** hostlist);
  */
 orte_ras_base_module_t orte_ras_loadleveler_module = {
     orte_ras_loadleveler_allocate,
-    orte_ras_base_node_insert,
-    orte_ras_base_node_query,
-    orte_ras_base_node_query_alloc,
-    orte_ras_base_node_lookup,
-    orte_ras_loadleveler_deallocate,
     orte_ras_loadleveler_finalize
 };
 
@@ -80,75 +67,55 @@ orte_ras_base_module_t orte_ras_loadleveler_module = {
  * Discover available (pre-allocated) nodes.  Allocate the
  * requested number of nodes/process slots to the job.
  */
-static int orte_ras_loadleveler_allocate(orte_jobid_t jobid, opal_list_t *attributes)
+static int orte_ras_loadleveler_allocate(opal_list_t *nodes)
 {
-    int i, ret;
-    opal_list_t nodes_list;
+    int i, ret=ORTE_SUCCESS;
     opal_list_item_t* item;
-    orte_ras_node_t* node;
+    orte_node_t* node;
     char ** hostlist = NULL;
     int num_hosts = 0;
 
-    OBJ_CONSTRUCT(&nodes_list, opal_list_t);
-
     ret = orte_ras_loadleveler_get_hostlist(&num_hosts, &hostlist);
     if(ORTE_SUCCESS != ret) {
-        goto cleanup;
+        ORTE_ERROR_LOG(ret);
+        return ret;
     }
 
     for (i = 0; i < num_hosts; i++) {
         /* check for duplicated nodes */
-        for (item = opal_list_get_first(&nodes_list); 
-             opal_list_get_end(&nodes_list) != item;
+        for (item = opal_list_get_first(nodes); 
+             opal_list_get_end(nodes) != item;
              item = opal_list_get_next(item)) {
-             node = (orte_ras_node_t*) item;
-             if (0 == strcmp(node->node_name, hostlist[i])) {
-                ++node->node_slots;
+             node = (orte_node_t*) item;
+             if (0 == strcmp(node->name, hostlist[i])) {
+                ++node->slots;
                 break;
             }
         }
      
-        if(opal_list_get_end(&nodes_list) == item) {
+        if(opal_list_get_end(nodes) == item) {
             /* we did not find a duplicate, so add a new item to the list */
-            node = OBJ_NEW(orte_ras_node_t);
+            node = OBJ_NEW(orte_node_t);
             if (NULL == node) {
+                ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
                 ret = ORTE_ERR_OUT_OF_RESOURCE;
                 goto cleanup;
             }
-            node->node_name = strdup(hostlist[i]);
-            node->node_arch = NULL;
-            node->node_state = ORTE_NODE_STATE_UP;
-            node->node_cellid = 0;
-            node->node_slots_inuse = 0;
-            node->node_slots_max = 0;
-            node->node_slots = 1;
-            opal_list_append(&nodes_list, &node->super);
+            node->name = strdup(hostlist[i]);
+            node->state = ORTE_NODE_STATE_UP;
+            node->slots_inuse = 0;
+            node->slots_max = 0;
+            node->slots = 1;
+            opal_list_append(nodes, &node->super);
         }   
-    }       
-    ret = orte_ras_base_node_insert(&nodes_list);
-    ret = orte_ras_base_allocate_nodes(jobid, &nodes_list);
-
-cleanup:            
-    while (NULL != (item = opal_list_remove_first(&nodes_list))) {
-        OBJ_RELEASE(item);    
-    }       
-    OBJ_DESTRUCT(&nodes_list);
+    }
+    
+cleanup:
     opal_argv_free(hostlist);
             
     return ret;
 }
 
-/*
- * There's really nothing to do here
- */
-static int orte_ras_loadleveler_deallocate(orte_jobid_t jobid)
-{
-    opal_output(orte_ras_base.ras_output, 
-                "ras:loadleveler:deallocate: success (nothing to do)");
-    return ORTE_SUCCESS;
-}
-
-
 /*
  * There's really nothing to do here
  */
@@ -176,7 +143,7 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     /* Get the step ID from LOADL_STEP_ID environment variable. */
     if(NULL == (ll_step_id = getenv("LOADL_STEP_ID"))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: could not get LOADL_STEP_ID "
+                    "ras:loadleveler:get:hostlist: could not get LOADL_STEP_ID "
                     "from environment!");
         return ORTE_ERROR;
     }
@@ -191,7 +158,7 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     /* Initialize the LL API. Specify that query type is JOBS. */
     if(NULL == (queryObject = ll_query(JOBS))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_query faild on JOBS!");
+                    "ras:loadleveler:get:hostlist: 1 ll_query faild on JOBS!");
         return ORTE_ERROR;
     }
 
@@ -199,8 +166,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     rc = ll_set_request(queryObject, QUERY_STEPID, job_step_list, ALL_DATA);
     if(0 > rc) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_set request failed: error "
-                    "%d!", rc);
+                    "ras:loadleveler:get:hostlist: 1 ll_set_request failed: "
+                    "error %d!", rc);
         return ORTE_ERROR;
     }
 
@@ -208,21 +175,22 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     job = ll_get_objs(queryObject, LL_CM, NULL, &obj_count, &err_code);
     if(NULL == job) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_objs faild!");
+                    "ras:loadleveler:get:hostlist: ll_get_objs LL_CM "
+                    "failed: err_code=%d", err_code);
         return ORTE_ERROR;
     }
 
     if (obj_count != 1) {  /* Only 1 Job object is expected. */
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_objs: expected one job "
-                    "to match, got %d!", obj_count);
+                    "ras:loadleveler:get:hostlist: ll_get_objs LL_CM "
+                    "expected one job to match, got %d!", obj_count);
         return ORTE_ERROR;
     }
 
     if(0 != (rc = ll_get_data(job, LL_JobSchedd, &schedd_host_name))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_data: failure. RC= %d!", 
-                    rc);
+                    "ras:loadleveler:get:hostlist: ll_get_data LL_JobSchedd"
+                    " failure, RC= %d!", rc);
         return ORTE_ERROR;
     }
     if (schedd_host_name != NULL) {
@@ -230,8 +198,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
         job_step_list[1] = NULL;
     } else {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_objs() Error: Could not "
-                    "determine managing schedd for job %s.\n",
+                    "ras:loadleveler:get:hostlist: ll_get_data() Error: Could "
+                    "not determine managing schedd for job %s.\n",
                     job_step_list[0]);
         return ORTE_ERROR;
     }
@@ -244,7 +212,7 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     /* Initialize the LL API. Specify that query type is JOBS. */
     if(NULL == (queryObject = ll_query(JOBS))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_query faild on JOBS!");
+                    "ras:loadleveler:get:hostlist: 2 ll_query faild on JOBS!");
         return ORTE_ERROR;
     }
 
@@ -252,8 +220,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     rc = ll_set_request(queryObject, QUERY_STEPID, job_step_list, ALL_DATA);
     if(0 != rc) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_set request failed: error "
-                    "%d!", rc);
+                    "ras:loadleveler:get:hostlist: 2 ll_set_request failed: "
+                    "error %d!", rc);
         return ORTE_ERROR;
     }
 
@@ -262,49 +230,49 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
                       &err_code);
     if(NULL == job) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_set request failed: error "
-                    "%d!", rc);
+                    "ras:loadleveler:get:hostlist: ll_get_objs LL_SCHEDD "
+                    "failed: err_code=%d", err_code);
         return ORTE_ERROR;
     }
 
     if (obj_count != 1) {  /* Only 1 Job object is expected. */
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_objs: expected one job " 
-                    "to match, got %d!", obj_count);
+                    "ras:loadleveler:get:hostlist: ll_get_objs LL_SCHEDD "
+                    "expected one job to match, got %d!", obj_count);
         return ORTE_ERROR;
     }
 
     if(0 != (rc = ll_get_data(job, LL_JobStepCount, &job_step_count))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_data: failure. RC= %d!",
-                    rc);
+                    "ras:loadleveler:get:hostlist: ll_get_data LL_JobStepCount"
+                    " failure, RC= %d!", rc);
         return ORTE_ERROR;
     }
     if (job_step_count != 1) { /* Only 1 Job Step object is expected. */
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_objs: expected one job "
-                    "step to match, got %d!", obj_count);
+                    "ras:loadleveler:get:hostlist: ll_get_data LL_JobStepCount"
+                    " expected one jobstep to match, got %d!", job_step_count);
         return ORTE_ERROR;
     }
 
     step = NULL;
     if(0 != (rc = ll_get_data(job, LL_JobGetFirstStep, &step))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_data: failure. RC= %d!",
-                    rc);
+                    "ras:loadleveler:get:hostlist: 3 ll_get_data: failure on "
+                    "LL_JobGetFirstStep. RC= %d!", rc);
         return ORTE_ERROR;
     }
     if(NULL == step) {
         opal_output(orte_ras_base.ras_output,
-                    "ll_get_data() Error: Unable to obtain Job Step "
-                    "information.\n");
+                    "ras:loadleveler:get:hostlist: 3 ll_get_data: Error: "
+                    "Unable to obtain Job Step information.\n");
         return ORTE_ERROR;
     }
 
     step_mode = -1;
     if(0 != (rc = ll_get_data(step, LL_StepParallelMode, &step_mode))) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: ll_get_data: failure on "
+                    "ras:loadleveler:get:hostlist: 4 ll_get_data: failure on "
                     "LL_StepParallelMode. RC= %d!", rc);
         return ORTE_ERROR;
     }
@@ -312,7 +280,7 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
     /* Serial job step: step_mode==0; Parallel: step_mode==1; Others:2,3,4. */
     if ((step_mode != 0) && (step_mode != 1)) {
         opal_output(orte_ras_base.ras_output,
-                    "ras:loadleveler:allocate: We support only Serial and "
+                    "ras:loadleveler:get:hostlist: We support only Serial and "
                     "Parallel LoadLeveler job types. PVM, NQS, and Blue Gene"
                     "jobs are not supported by the LoadLeveler RAS!");
         return ORTE_ERROR;
@@ -322,31 +290,31 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
         node = NULL;
         if(0 != (rc = ll_get_data(step, LL_StepGetFirstNode, &node))) {
             opal_output(orte_ras_base.ras_output,
-                        "ras:loadleveler:allocate: ll_get_data: failure on "
-                        "LL_StepGetFirstNode. RC= %d!", rc);
+                        "ras:loadleveler:get:hostlist: ll_get_data: failure "
+                        "on serial LL_StepGetFirstNode. RC= %d!", rc);
             return ORTE_ERROR;
         }
         task = NULL;
         if(0 != (rc = ll_get_data(node, LL_NodeGetFirstTask, &task))) {
             opal_output(orte_ras_base.ras_output,
-                        "ras:loadleveler:allocate: ll_get_data: failure on "
-                        "LL_NodeGetFirstTask. RC= %d!", rc);
+                        "ras:loadleveler:get:hostlist: ll_get_data: failure "
+                        "on serial LL_NodeGetFirstTask. RC= %d!", rc);
             return ORTE_ERROR;
         }
         task_instance = NULL;
         rc = ll_get_data(task, LL_TaskGetFirstTaskInstance, &task_instance);
         if(0 != rc) {
             opal_output(orte_ras_base.ras_output,
-                        "ras:loadleveler:allocate: ll_get_data: failure on "
-                        "LL_TaskGetFirstInstance. RC= %d!", rc);
+                        "ras:loadleveler:get:hostlist: ll_get_data: failure "
+                        "on serial LL_TaskGetFirstInstance. RC= %d!", rc);
             return ORTE_ERROR;
         }
         task_machine_name = NULL;
         if(0 != (rc = ll_get_data(task_instance, LL_TaskInstanceMachineName, 
                                   &task_machine_name))) {
             opal_output(orte_ras_base.ras_output,
-                        "ras:loadleveler:allocate: ll_get_data: failure on " 
-                        "LL_TaskInstanceMachineName. RC= %d!", rc);
+                        "ras:loadleveler:get:hostlist: ll_get_data: failure "
+                        "on serial LL_TaskInstanceMachineName. RC= %d!", rc);
             return ORTE_ERROR;
         }
         opal_argv_append(num_hosts, hostlist, task_machine_name);
@@ -356,8 +324,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
         node = NULL;
         if(0 != (rc = ll_get_data(step, LL_StepGetFirstNode, &node))) {
             opal_output(orte_ras_base.ras_output,
-                        "ras:loadleveler:allocate: ll_get_data: failure on "
-                        "LL_StepGetFirstNode. RC= %d!", rc);
+                        "ras:loadleveler:get:hostlist: ll_get_data: failure "
+                        "on LL_StepGetFirstNode. RC= %d!", rc);
             return ORTE_ERROR;
         }
     
@@ -365,8 +333,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
             task = NULL;
             if(0 != (rc = ll_get_data(node, LL_NodeGetFirstTask, &task))) {
                 opal_output(orte_ras_base.ras_output,
-                            "ras:loadleveler:allocate: ll_get_data: failure on "
-                            "LL_NodeGetFirstTask. RC= %d!", rc);
+                            "ras:loadleveler:get:hostlist: ll_get_data: failure "
+                            "on LL_NodeGetFirstTask. RC= %d!", rc);
                 return ORTE_ERROR;
             }
         
@@ -375,8 +343,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
                 rc = ll_get_data(task, LL_TaskIsMaster, &ll_master_task);
                 if(0 != rc) {
                     opal_output(orte_ras_base.ras_output,
-                                "ras:loadleveler:allocate: ll_get_data: failure" 
-                                "  on LL_TaskIsMaster. RC= %d!", rc);
+                                "ras:loadleveler:get:hostlist: ll_get_data: " 
+                                "failure on LL_TaskIsMaster. RC= %d!", rc);
                     return ORTE_ERROR;
                 }
             
@@ -388,8 +356,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
                     if(0 != (rc = ll_get_data(task, LL_TaskGetFirstTaskInstance,
                                               &task_instance))) {
                         opal_output(orte_ras_base.ras_output,
-                                    "ras:loadleveler:allocate: ll_get_data: "
-                                    "failure on LL_TaskGetFirstTaskInstance. "
+                                    "ras:loadleveler:get:hostlist: ll_get_data:"
+                                    " failure on LL_TaskGetFirstTaskInstance."
                                     " RC= %d!", rc);
                         return ORTE_ERROR;
                     }
@@ -402,7 +370,7 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
                                          &task_machine_name);
                         if(0 != rc) {
                             opal_output(orte_ras_base.ras_output,
-                                        "ras:loadleveler:allocate: ll_get_data:"
+                                        "ras:loadleveler:get:hostlist: ll_get_data:"
                                         " failure on LL_TaskInstanceMachineName"
                                         "RC= %d!", rc);
                             return ORTE_ERROR;
@@ -413,8 +381,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
                                          &task_instance);
                         if(0 != rc) {
                             opal_output(orte_ras_base.ras_output,
-                                        "ras:loadleveler:allocate: ll_get_data:"
-                                        " failure on LL_TaskGetNextInstance. "
+                                        "ras:loadleveler:get:hostlist: ll_get_data:"
+                                        " failure on LL_TaskGetNextTaskInstance. "
                                         "RC= %d!", rc);
                             return ORTE_ERROR;
                         }
@@ -423,7 +391,7 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
                 task = NULL;
                 if(0 != (rc = ll_get_data(node, LL_NodeGetNextTask, &task))) {
                     opal_output(orte_ras_base.ras_output,
-                                "ras:loadleveler:allocate: ll_get_data: " 
+                                "ras:loadleveler:get:hostlist: ll_get_data: " 
                                 "failure on LL_NodeGetNextTask. RC= %d!", rc);
                     return ORTE_ERROR;
                 }
@@ -431,8 +399,8 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
             node = NULL;
             if(0 != (rc = ll_get_data(step, LL_StepGetNextNode, &node))) {
                 opal_output(orte_ras_base.ras_output,
-                            "ras:loadleveler:allocate: ll_get_data: failure "
-                            "on LL_StepGetNextNode. RC= %d!", rc);
+                            "ras:loadleveler:get:hostlist: ll_get_data: "
+                            "failure on LL_StepGetNextNode. RC= %d!", rc);
                 return ORTE_ERROR;
             }
         }
@@ -442,55 +410,3 @@ static int orte_ras_loadleveler_get_hostlist(int* num_hosts, char*** hostlist)
 
     return ORTE_SUCCESS;
 }
-
-#if 0
-/* For now, we do not get the node architectures from LoadLeveler. It is slow,
- * and we don't even use the value. */
-/*
- * get the machine arch from LoadLeveler
- * Will return NULL on error or a arch string that needs to be freed
- * (some code from the IBM documentation, licensed as above)
- */
-static char* orte_ras_loadleveler_get_host_arch(char * hostname) {
-    LL_element *queryObject, *machine; 
-    int rc, obj_count, err_code;
-    char * hostlist[2];
-    char * arch;
-  
-    /* Initialize the query: Machine query */
-    queryObject = ll_query(MACHINES);
-    if(NULL == queryObject) {
-        return NULL;
-    } 
-  
-    /* Set query parameters: query specific machines by name */ 
-    hostlist[0] = hostname;
-    hostlist[1] = NULL; 
-   
-    rc = ll_set_request(queryObject, QUERY_HOST, hostlist, ALL_DATA); 
-    if(0 != rc) { 
-        return NULL;
-    } 
-  
-    /* Get the machine objects from the LoadL_negotiator (central manager) daemon */ 
-    machine = ll_get_objs(queryObject, LL_CM, NULL, &obj_count, &err_code); 
-    if(NULL == machine || 1 != obj_count) {
-        return NULL;
-    } 
- 
-    /* Process the machine object */
-    rc = ll_get_data(machine, LL_MachineArchitecture, &arch); 
-    if(0 != rc) { 
-        return NULL;
-    } 
-  
-    /* Free objects obtained from Negotiator */ 
-    ll_free_objs(queryObject); 
- 
-    /* Free query element */ 
-    ll_deallocate(queryObject); 
-    
-    return arch;
-}
-#endif
-
