@@ -13,7 +13,7 @@
  * Copyright (c) 2007      Sun Microsystems, Inc. All rights reserved.
  * Copyright (c) 2007      Los Alamos National Security, LLC.  All rights
  *                         reserved. 
- * Copyright (c) 2007-2009 Cisco, Inc.  All rights reserved.
+ * Copyright (c) 2007-2010 Cisco Systems, Inc.  All rights reserved.
  * $COPYRIGHT$
  * 
  * Additional copyrights may follow
@@ -92,6 +92,9 @@
  * hard-coded in the OMPI layer (see ompi/debuggers/ompi_debuggers.c).
  */
 
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
 #include <stdio.h>
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
@@ -105,20 +108,31 @@
 #include <ctype.h>
 
 #include "opal/util/argv.h"
+#include "opal/util/output.h"
 #include "opal/util/path.h"
 #include "opal/util/os_path.h"
+#include "opal/class/opal_pointer_array.h"
 #include "opal/mca/base/mca_base_param.h"
+#include "opal/util/opal_getcwd.h"
 
 #include "orte/mca/errmgr/errmgr.h"
 #include "orte/mca/rml/rml.h"
+#include "orte/mca/rml/rml_types.h"
+#include "orte/mca/plm/plm.h"
+#include "orte/mca/plm/base/plm_private.h"
+#include "orte/mca/rmaps/rmaps_types.h"
 #include "orte/runtime/orte_globals.h"
 #include "orte/runtime/orte_wait.h"
 #include "orte/util/show_help.h"
+#include "orte/util/name_fns.h"
 
 #include "debuggers.h"
 
 /* +++ begin MPICH/TotalView std debugger interface definitions */
 
+#define MPIR_MAX_PATH_LENGTH 256
+#define MPIR_MAX_ARG_LENGTH 1024
+
 struct MPIR_PROCDESC {
     char *host_name;        /* something that can be passed to inet_addr */
     char *executable_name;  /* name of binary */
@@ -127,10 +141,14 @@ struct MPIR_PROCDESC {
 
 struct MPIR_PROCDESC *MPIR_proctable = NULL;
 int MPIR_proctable_size = 0;
-int MPIR_being_debugged = 0;
+volatile int MPIR_being_debugged = 0;
 volatile int MPIR_debug_state = 0;
 volatile int MPIR_i_am_starter = 0;
 volatile int MPIR_partial_attach_ok = 1;
+volatile char MPIR_executable_path[MPIR_MAX_PATH_LENGTH];
+volatile char MPIR_server_arguments[MPIR_MAX_ARG_LENGTH];
+volatile int MPIR_forward_output = 0;
+volatile int MPIR_forward_comm = 0;
 
 /* --- end MPICH/TotalView std debugger interface definitions */
 
@@ -145,6 +163,7 @@ static void dump(void)
     DUMP_INT(MPIR_debug_state);
     DUMP_INT(MPIR_partial_attach_ok);
     DUMP_INT(MPIR_i_am_starter);
+    DUMP_INT(MPIR_forward_output);
     DUMP_INT(MPIR_proctable_size);
     fprintf(stderr, "  MPIR_proctable:\n");
     for (i = 0; i < MPIR_proctable_size; i++) {
@@ -155,8 +174,15 @@ static void dump(void)
                 MPIR_proctable[i].executable_name,
                 MPIR_proctable[i].pid);
     }
+    fprintf(stderr, "MPIR_executable_path: %s\n",
+            ('\0' == MPIR_executable_path[0]) ?
+            "NULL" : (char*) MPIR_executable_path);
+    fprintf(stderr, "MPIR_server_arguments: %s\n",
+            ('\0' == MPIR_server_arguments[0]) ?
+            "NULL" : (char*) MPIR_server_arguments);
 }
 
+
 /*
  * Process one line from the orte_base_user_debugger MCA param and
  * look for that debugger in the path.  If we find it, fill in
@@ -165,17 +191,19 @@ static void dump(void)
 static int process(char *orig_line, char *basename, opal_cmd_line_t *cmd_line,
                    int argc, char **argv, char ***new_argv, int num_procs) 
 {
-    int i;
-    char *line, *full_line = strdup(orig_line);
-    char *user_argv, *tmp, *tmp2, **tmp_argv, **executable;
-    char cwd[OMPI_PATH_MAX];
+    int ret = ORTE_SUCCESS;
+    int i, j, count;
+    char *line = NULL, *tmp = NULL, *full_line = strdup(orig_line);
+    char **orterun_argv = NULL, **executable_argv = NULL, **line_argv = NULL;
+    char cwd[OPAL_PATH_MAX];
     bool used_num_procs = false;
     bool single_app = false;
     bool fail_needed_executable = false;
 
     line = full_line;
     if (NULL == line) {
-        return ORTE_ERR_OUT_OF_RESOURCE;
+        ret = ORTE_ERR_OUT_OF_RESOURCE;
+        goto out;
     }
 
     /* Trim off whitespace at the beginning and ending of line */
@@ -187,117 +215,89 @@ static int process(char *orig_line, char *basename, opal_cmd_line_t *cmd_line,
         line[i] = '\0';
     }
     if (strlen(line) <= 0) {
-        return ORTE_ERROR;
+        ret = ORTE_ERROR;
+        goto out;
     }
 
     /* Get the tail of the command line (i.e., the user executable /
        argv) */
 
-    opal_cmd_line_get_tail(cmd_line, &i, &executable);
+    opal_cmd_line_get_tail(cmd_line, &i, &executable_argv);
 
-    /* Remove --debug, --debugger, and -tv from the user command line
-       params */
+    /* Make a new copy of the orterun command line args, without the
+       orterun token itself, and without the --debug, --debugger, and
+       -tv flags. */
 
-    if (1 == argc) {
-        user_argv = strdup("");
-    } else {
-        tmp_argv = opal_argv_copy(argv);
-        for (i = 0; NULL != tmp_argv[i]; ++i) {
-            if (0 == strcmp(tmp_argv[i], "-debug") ||
-                0 == strcmp(tmp_argv[i], "--debug")) {
-                free(tmp_argv[i]);
-                tmp_argv[i] = strdup("");
-            } else if (0 == strcmp(tmp_argv[i], "-tv") ||
-                0 == strcmp(tmp_argv[i], "--tv")) {
-                free(tmp_argv[i]);
-                tmp_argv[i] = strdup("");
-            } else if (0 == strcmp(tmp_argv[i], "--debugger") ||
-                       0 == strcmp(tmp_argv[i], "-debugger")) {
-                free(tmp_argv[i]);
-                tmp_argv[i] = strdup("");
-                if (NULL != tmp_argv[i + 1]) {
-                    ++i;
-                    free(tmp_argv[i]);
-                    tmp_argv[i] = strdup("");
-                }
-            }
+    orterun_argv = opal_argv_copy(argv);
+    count = opal_argv_count(orterun_argv);
+    opal_argv_delete(&count, &orterun_argv, 0, 1);
+    for (i = 0; NULL != orterun_argv[i]; ++i) {
+        count = opal_argv_count(orterun_argv);
+        if (0 == strcmp(orterun_argv[i], "-debug") ||
+            0 == strcmp(orterun_argv[i], "--debug")) {
+            opal_argv_delete(&count, &orterun_argv, i, 1);
+        } else if (0 == strcmp(orterun_argv[i], "-tv") ||
+                   0 == strcmp(orterun_argv[i], "--tv")) {
+            opal_argv_delete(&count, &orterun_argv, i, 1);
+        } else if (0 == strcmp(orterun_argv[i], "--debugger") ||
+                   0 == strcmp(orterun_argv[i], "-debugger")) {
+            opal_argv_delete(&count, &orterun_argv, i, 2);
         }
-        user_argv = opal_argv_join(tmp_argv + 1, ' ');
-        opal_argv_free(tmp_argv);
     }
 
     /* Replace @@ tokens - line should never realistically be bigger
        than MAX_INT, so just cast to int to remove compiler warning */
 
-    for (i = 0; i < (int) strlen(line); ++i) {
-        tmp = NULL;
-        if (0 == strncmp(line + i, "@mpirun@", 8)) {
-            line[i] = '\0';
-            asprintf(&tmp, "%s%s%s", line, argv[0], line + i + 8);
-        } else if (0 == strncmp(line + i, "@orterun@", 9)) {
-            line[i] = '\0';
-            asprintf(&tmp, "%s%s%s", line, argv[0], line + i + 9);
-        } else if (0 == strncmp(line + i, "@mpirun_args@", 13)) {
-            line[i] = '\0';
-            asprintf(&tmp, "%s%s%s", line, user_argv, line + i + 13);
-        } else if (0 == strncmp(line + i, "@orterun_args@", 14)) {
-            line[i] = '\0';
-            asprintf(&tmp, "%s%s%s", line, user_argv, line + i + 14);
-        } else if (0 == strncmp(line + i, "@np@", 4)) {
-            line[i] = '\0';
-            asprintf(&tmp, "%s%d%s", line, num_procs,
-                     line + i + 4);
-            used_num_procs = true;
-        } else if (0 == strncmp(line + i, "@single_app@", 12)) {
-            line[i] = '\0';
+    *new_argv = NULL;
+    line_argv = opal_argv_split(line, ' ');
+    if (NULL == line_argv) {
+        ret = ORTE_ERR_NOT_FOUND;
+        goto out;
+    }
+    for (i = 0; NULL != line_argv[i]; ++i) {
+        if (0 == strcmp(line_argv[i], "@mpirun@") ||
+            0 == strcmp(line_argv[i], "@orterun@")) {
+            opal_argv_append_nosize(new_argv, argv[0]);
+        } else if (0 == strcmp(line_argv[i], "@mpirun_args@") ||
+                   0 == strcmp(line_argv[i], "@orterun_args@")) {
+            for (j = 0; NULL != orterun_argv && NULL != orterun_argv[j]; ++j) {
+                opal_argv_append_nosize(new_argv, orterun_argv[j]);
+            }
+        } else if (0 == strcmp(line_argv[i], "@np@")) {
+            asprintf(&tmp, "%d", num_procs);
+            opal_argv_append_nosize(new_argv, tmp);
+            free(tmp);
+        } else if (0 == strcmp(line_argv[i], "@single_app@")) {
             /* This token is only a flag; it is not replaced with any
                alternate text */
-            asprintf(&tmp, "%s%s", line, line + i + 12);
             single_app = true;
-        } else if (0 == strncmp(line + i, "@executable@", 12)) {
-            line[i] = '\0';
+        } else if (0 == strcmp(line_argv[i], "@executable@")) {
             /* If we found the executable, paste it in.  Otherwise,
                this is a possible error. */
-            if (NULL != executable) {
-                asprintf(&tmp, "%s%s%s", line, executable[0], line + i + 12);
+            if (NULL != executable_argv) {
+                opal_argv_append_nosize(new_argv, executable_argv[0]);
             } else {
                 fail_needed_executable = true;
             }
-        } else if (0 == strncmp(line + i, "@executable_argv@", 17)) {
-            line[i] = '\0';
+        } else if (0 == strcmp(line_argv[i], "@executable_argv@")) {
             /* If we found the tail, paste in the argv.  Otherwise,
                this is a possible error. */
-            if (NULL != executable) {
-                if (NULL != executable[1]) {
-                    /* Put in the argv */
-                    tmp2 = opal_argv_join(executable + 1, ' ');
-                    asprintf(&tmp, "%s%s%s", line, tmp2, line + i + 17);
-                    free(tmp2);
-                } else {
-                    /* There is no argv; just paste the front and back
-                       together, removing the @token@ */
-                    asprintf(&tmp, "%s%s", line, line + i + 17);
+            if (NULL != executable_argv) {
+                for (j = 1; NULL != executable_argv[j]; ++j) {
+                    opal_argv_append_nosize(new_argv, executable_argv[j]);
                 }
             } else {
                 fail_needed_executable = true;
             }
-        }
-
-        if (NULL != tmp) {
-            free(full_line);
-            full_line = line = tmp;
-            --i;
+        } else {
+            /* It wasn't a special token, so just copy it over */
+            opal_argv_append_nosize(new_argv, line_argv[i]);
         }
     }
 
-    /* Split up into argv */
-
-    *new_argv = opal_argv_split(line, ' ');
-    free(full_line);
-
     /* Can we find argv[0] in the path? */
 
-    getcwd(cwd, OMPI_PATH_MAX);
+    getcwd(cwd, OPAL_PATH_MAX);
     tmp = opal_path_findv((*new_argv)[0], X_OK, environ, cwd);
     if (NULL != tmp) {
         free(tmp);
@@ -310,8 +310,10 @@ static int process(char *orig_line, char *basename, opal_cmd_line_t *cmd_line,
            -np value if the user did not specify -np on the command
            line. */
         if (used_num_procs && 0 == num_procs) {
+            free(tmp);
+            tmp = opal_argv_join(orterun_argv, ' ');
             orte_show_help("help-orterun.txt", "debugger requires -np",
-                           true, (*new_argv)[0], argv[0], user_argv, 
+                           true, (*new_argv)[0], argv[0], tmp,
                            (*new_argv)[0]);
             /* Fall through to free / fail, below */
         } 
@@ -337,8 +339,7 @@ static int process(char *orig_line, char *basename, opal_cmd_line_t *cmd_line,
 
         /* Otherwise, we succeeded.  Return happiness. */
         else {
-            free(tmp);
-            return ORTE_SUCCESS;
+            goto out;
         }
         free(tmp);
     }
@@ -347,7 +348,25 @@ static int process(char *orig_line, char *basename, opal_cmd_line_t *cmd_line,
 
     opal_argv_free(*new_argv);
     *new_argv = NULL;
-    return ORTE_ERR_NOT_FOUND;
+    ret = ORTE_ERR_NOT_FOUND;
+
+ out:
+    if (NULL != orterun_argv) {
+        opal_argv_free(orterun_argv);
+    }
+    if (NULL != executable_argv) {
+        opal_argv_free(executable_argv);
+    }
+    if (NULL != line_argv) {
+        opal_argv_free(line_argv);
+    }
+    if (NULL != tmp) {
+        free(tmp);
+    }
+    if (NULL != full_line) {
+        free(full_line);
+    }
+    return ret;
 }
 
 /**
@@ -397,6 +416,11 @@ void orte_run_debugger(char *basename, opal_cmd_line_t *cmd_line,
     opal_argv_free(lines);
 
     /* We found one */
+    
+    /* cleanup the MPIR arrays in case the debugger doesn't set them */
+    memset((char*)MPIR_executable_path, 0, MPIR_MAX_PATH_LENGTH);
+    memset((char*)MPIR_server_arguments, 0, MPIR_MAX_ARG_LENGTH);
+    
     /* Set an MCA param so that everyone knows that they are being
        launched under a debugger; not all debuggers are consistent
        about setting MPIR_being_debugged in both the launcher and the
@@ -419,6 +443,107 @@ void orte_run_debugger(char *basename, opal_cmd_line_t *cmd_line,
 }
 
 
+static void build_debugger_args(orte_app_context_t *debugger)
+{
+    int i, j;
+    char mpir_arg[MPIR_MAX_ARG_LENGTH];
+
+    if ('\0' != MPIR_server_arguments[0]) {
+        j=0;
+        memset(mpir_arg, 0, MPIR_MAX_ARG_LENGTH);
+        for (i=0; i < MPIR_MAX_ARG_LENGTH; i++) {
+            if (MPIR_server_arguments[i] == '\0') {
+                if (0 < j) {
+                    opal_argv_append_nosize(&debugger->argv, mpir_arg);
+                    memset(mpir_arg, 0, MPIR_MAX_ARG_LENGTH);
+                    j=0;
+                }
+            } else {
+                mpir_arg[j] = MPIR_server_arguments[i];
+                j++;
+            }
+        }
+    }
+}
+
+static void check_debugger(int fd, short event, void *arg)
+{
+    struct timeval now;
+    opal_event_t *tmp = (opal_event_t*)arg;
+    orte_job_t *jdata;
+    orte_app_context_t *app;
+    char cwd[OPAL_PATH_MAX];
+    int rc;
+    int32_t ljob;
+
+    if (MPIR_being_debugged) {
+        if (orte_debug_flag) {
+            opal_output(0, "%s Launching debugger %s", ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
+            MPIR_executable_path);
+        }
+        
+        /* a debugger has attached! All the MPIR_Proctable
+         * data is already available, so we only need to
+         * check to see if we should spawn any daemons
+         */
+        if ('\0' != MPIR_executable_path[0]) {
+            /* this will be launched just like a regular job,
+             * so we do not use the global orte_debugger_daemon
+             * as this is reserved for co-location upon startup
+             */
+            jdata = OBJ_NEW(orte_job_t);
+            /* create a jobid for these daemons - this is done solely
+             * to avoid confusing the rest of the system's bookkeeping
+             */
+            orte_plm_base_create_jobid(jdata);
+            /* flag the job as being debugger daemons */
+            jdata->controls |= ORTE_JOB_CONTROL_DEBUGGER_DAEMON;
+            /* unless directed, we do not forward output */
+            if (!MPIR_forward_output) {
+                jdata->controls &= ~ORTE_JOB_CONTROL_FORWARD_OUTPUT;
+            }
+            /* set the mapping policy to "pernode" so we only get
+             * one debugger daemon on each node
+             */
+            jdata->map = OBJ_NEW(orte_job_map_t);
+            jdata->map->npernode = 1;
+            /* add it to the global job pool */
+            ljob = ORTE_LOCAL_JOBID(jdata->jobid);
+            opal_pointer_array_set_item(orte_job_data, ljob, jdata);
+            /* create an app_context for the debugger daemon */
+            app = OBJ_NEW(orte_app_context_t);
+            app->app = strdup((char*)MPIR_executable_path);
+            if (OPAL_SUCCESS != (rc = opal_getcwd(cwd, sizeof(cwd)))) {
+                orte_show_help("help-orterun.txt", "orterun:init-failure",
+                               true, "get the cwd", rc);
+                OBJ_RELEASE(jdata);
+                goto RELEASE;
+            }
+            app->cwd = strdup(cwd);
+            app->user_specified_cwd = false;
+            opal_argv_append_nosize(&app->argv, app->app);
+            build_debugger_args(app);
+            opal_pointer_array_add(jdata->apps, &app->super);
+            jdata->num_apps = 1;
+            /* now go ahead and spawn this job */
+            if (ORTE_SUCCESS != (rc = orte_plm.spawn(jdata))) {
+                ORTE_ERROR_LOG(rc);
+            }
+        }
+        
+    RELEASE:
+        /* notify the debugger that all is ready */
+        MPIR_Breakpoint();
+
+    } else {
+        /* reissue the timer to wake us up again */
+        now.tv_sec = orte_debugger_check_rate;
+        now.tv_usec = 0;
+        opal_evtimer_add(tmp, &now);
+    }
+}
+
+
 /**
  * Initialization of data structures for running under a debugger
  * using the MPICH/TotalView parallel debugger interface.  Before the
@@ -428,14 +553,23 @@ void orte_run_debugger(char *basename, opal_cmd_line_t *cmd_line,
 void orte_debugger_init_before_spawn(orte_job_t *jdata)
 {
     char *env_name;
-    orte_app_context_t **apps;
+    orte_app_context_t **apps, *app;
     orte_std_cntr_t i;
+    int32_t ljob;
 
     if (!MPIR_being_debugged && !orte_in_parallel_debugger) {
-        /* not being debugged */
+        /* not being debugged - check if we want to enable
+         * later attachment by debugger
+         */
+        if (orte_enable_debug_cospawn_while_running) {
+            /* setup a timer to wake us up periodically
+             * to check for debugger attach
+             */
+            ORTE_TIMER_EVENT(orte_debugger_check_rate, 0, check_debugger);
+        }
         return;
     }
-
+    
     if (orte_debug_flag) {
         opal_output(0, "Info: Spawned by a debugger");
     }
@@ -449,6 +583,32 @@ void orte_debugger_init_before_spawn(orte_job_t *jdata)
         opal_setenv(env_name, "1", true, &apps[i]->env);
     }
     free(env_name);
+    
+    /* check if we need to co-spawn the debugger daemons */
+    if ('\0' != MPIR_executable_path[0]) {
+        /* add debugger info to launch message */
+        orte_debugger_daemon = OBJ_NEW(orte_job_t);
+        /* create a jobid for these daemons - this is done solely
+         * to avoid confusing the rest of the system's bookkeeping
+         */
+        orte_plm_base_create_jobid(orte_debugger_daemon);
+        /* flag the job as being debugger daemons */
+        orte_debugger_daemon->controls |= ORTE_JOB_CONTROL_DEBUGGER_DAEMON;
+        /* unless directed, we do not forward output */
+        if (!MPIR_forward_output) {
+            orte_debugger_daemon->controls &= ~ORTE_JOB_CONTROL_FORWARD_OUTPUT;
+        }
+        /* add it to the global job pool */
+        ljob = ORTE_LOCAL_JOBID(orte_debugger_daemon->jobid);
+        opal_pointer_array_set_item(orte_job_data, ljob, orte_debugger_daemon);
+        /* create an app_context for the debugger daemon */
+        app = OBJ_NEW(orte_app_context_t);
+        app->app = strdup((char*)MPIR_executable_path);
+        opal_argv_append_nosize(&app->argv, app->app);
+        build_debugger_args(app);
+        opal_pointer_array_add(orte_debugger_daemon->apps, &app->super);
+        orte_debugger_daemon->num_apps = 1;
+    }
 }
 
 
@@ -463,8 +623,8 @@ void orte_debugger_init_before_spawn(orte_job_t *jdata)
  */
 void orte_debugger_init_after_spawn(orte_job_t *jdata)
 {
-    orte_proc_t **procs;
-    orte_app_context_t *appctx, **apps;
+    orte_proc_t *proc;
+    orte_app_context_t *appctx;
     orte_vpid_t i, j;
     opal_buffer_t buf;
     orte_process_name_t rank0;
@@ -495,19 +655,19 @@ void orte_debugger_init_after_spawn(orte_job_t *jdata)
     }
     
     /* initialize MPIR_proctable */
-    procs = (orte_proc_t**)jdata->procs->addr;
-    apps = (orte_app_context_t**)jdata->apps->addr;
     for (j=0; j < jdata->num_procs; j++) {
-        if (NULL == procs[j]) {
-            opal_output(0, "Error: undefined proc at position %ld\n", (long)j);
+        if (NULL == (proc = (orte_proc_t*)opal_pointer_array_get_item(jdata->procs, j))) {
+            continue;
         }
         /* store this data in the location whose index
          * corresponds to the proc's rank
          */
-        i = procs[j]->name.vpid;
-        appctx = apps[procs[j]->app_idx];
+        i = proc->name.vpid;
+        if (NULL == (appctx = (orte_app_context_t*)opal_pointer_array_get_item(jdata->apps, proc->app_idx))) {
+            continue;
+        }
         
-        MPIR_proctable[i].host_name = strdup(procs[j]->node->name);
+        MPIR_proctable[i].host_name = strdup(proc->node->name);
         if ( 0 == strncmp(appctx->app, OPAL_PATH_SEP, 1 )) { 
             MPIR_proctable[i].executable_name = 
             opal_os_path( false, appctx->app, NULL ); 
@@ -515,7 +675,7 @@ void orte_debugger_init_after_spawn(orte_job_t *jdata)
             MPIR_proctable[i].executable_name =
             opal_os_path( false, appctx->cwd, appctx->app, NULL ); 
         } 
-        MPIR_proctable[i].pid = procs[j]->pid;
+        MPIR_proctable[i].pid = proc->pid;
     }
 
     if (orte_debug_flag) {
