@@ -26,12 +26,14 @@
 
 #include "opal/util/show_help.h"
 #include "opal/util/output.h"
-#include "opal/util/printf.h"
 #include "opal/dss/dss.h"
 
 #include "orte/mca/errmgr/errmgr.h"
 #include "orte/mca/rml/rml.h"
+#include "orte/mca/rml/rml_types.h"
+#include "orte/mca/routed/routed.h"
 #include "orte/util/name_fns.h"
+#include "orte/util/proc_info.h"
 #include "orte/runtime/orte_globals.h"
 
 #include "orte/util/show_help.h"
@@ -116,6 +118,8 @@ static bool show_help_timer_set = false;
 static opal_event_t show_help_timer_event;
 static bool ready;
 
+static opal_show_help_fn_t save_help = NULL;
+
 static void tuple_list_item_constructor(tuple_list_item_t *obj)
 {
     obj->tli_filename = NULL;
@@ -144,6 +148,117 @@ static void tuple_list_item_destructor(tuple_list_item_t *obj)
     }
 }
 
+/* dealing with special characters in xml output */
+static char* xml_format(unsigned char *input)
+{
+    int i, j, k, len, outlen;
+    char *output, qprint[10];
+    char *endtag="</stderr>";
+    char *starttag="<stderr>";
+    int endtaglen, starttaglen;
+    bool endtagged = false;
+    
+    len = strlen((char*)input);
+    /* add some arbitrary size padding */
+    output = (char*)malloc((len+1024)*sizeof(char));
+    if (NULL == output) {
+        ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+        return (char*)input; /* default to no xml formatting */
+    }
+    memset(output, 0, len+1024);
+    outlen = len+1023;
+    endtaglen = strlen(endtag);
+    starttaglen = strlen(starttag);
+    
+    /* start at the beginning */
+    k=0;
+    
+    /* start with the tag */
+    for (j=0; j < starttaglen && k < outlen; j++) {
+        output[k++] = starttag[j];
+    }        
+
+    for (i=0; i < len; i++) {
+        if ('&' == input[i]) {
+            if (k+5 >= outlen) {
+                ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+                goto error;
+            }
+            snprintf(qprint, 10, "&amp;");
+            for (j=0; j < (int)strlen(qprint) && k < outlen; j++) {
+                output[k++] = qprint[j];
+            }
+        } else if ('<' == input[i]) {
+            if (k+4 >= outlen) {
+                ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+                goto error;
+            }
+            snprintf(qprint, 10, "&lt;");
+            for (j=0; j < (int)strlen(qprint) && k < outlen; j++) {
+                output[k++] = qprint[j];
+            }
+        } else if ('>' == input[i]) {
+            if (k+4 >= outlen) {
+                ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+                goto error;
+            }
+            snprintf(qprint, 10, "&gt;");
+            for (j=0; j < (int)strlen(qprint) && k < outlen; j++) {
+                output[k++] = qprint[j];
+            }
+        } else if (input[i] < 32 || input[i] > 127) {
+            /* this is a non-printable character, so escape it too */
+            if (k+7 >= outlen) {
+                ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+                goto error;
+            }
+            snprintf(qprint, 10, "&#%03d;", (int)input[i]);
+            for (j=0; j < (int)strlen(qprint) && k < outlen; j++) {
+                output[k++] = qprint[j];
+            }
+            /* if this was a \n, then we also need to break the line with the end tag */
+            if ('\n' == input[i] && (k+endtaglen+1) < outlen) {
+                /* we need to break the line with the end tag */
+                for (j=0; j < endtaglen && k < outlen-1; j++) {
+                    output[k++] = endtag[j];
+                }
+                /* move the <cr> over */
+                output[k++] = '\n';
+                /* if this isn't the end of the input buffer, add a new start tag */
+                if (i < len-1 && (k+starttaglen) < outlen) {
+                    for (j=0; j < starttaglen && k < outlen; j++) {
+                        output[k++] = starttag[j];
+                        endtagged = false;
+                    }
+                } else {
+                    endtagged = true;
+                }
+            }
+        } else {
+            output[k++] = input[i];
+        }    
+    }
+
+    if (!endtagged) {
+        /* need to add an endtag */
+        for (j=0; j < endtaglen && k < outlen-1; j++) {
+            output[k++] = endtag[j];
+        }
+        output[k++] = '\n';
+    }
+    
+    return output;
+    
+error:
+    /* if we couldn't complete the processing for
+     * some reason, return the unprocessed input
+     * so at least the message gets out!
+     */
+    free(output);
+    return (char*)input;
+}
+
+
 /*
  * Check to see if a given (filename, topic) tuple has been displayed
  * already.  Return ORTE_SUCCESS if so, or ORTE_ERR_NOT_FOUND if not.
@@ -191,6 +306,7 @@ static void show_accumulated_duplicates(int fd, short event, void *context)
     opal_list_item_t *item;
     time_t now = time(NULL);
     tuple_list_item_t *tli;
+    char *tmp, *output;
 
     /* Loop through all the messages we've displayed and see if any
        processes have sent duplicates that have not yet been displayed
@@ -201,14 +317,30 @@ static void show_accumulated_duplicates(int fd, short event, void *context)
         tli = (tuple_list_item_t*) item;
         if (tli->tli_count_since_last_display > 0) {
             static bool first = true;
-            opal_output(0, "%d more process%s sent help message %s / %s",
-                        tli->tli_count_since_last_display,
-                        (tli->tli_count_since_last_display > 1) ? "es have" : " has",
-                        tli->tli_filename, tli->tli_topic);
+            if (orte_xml_output) {
+                asprintf(&tmp, "%d more process%s sent help message %s / %s",
+                         tli->tli_count_since_last_display,
+                         (tli->tli_count_since_last_display > 1) ? "es have" : " has",
+                         tli->tli_filename, tli->tli_topic);
+                output = xml_format((unsigned char*)tmp);
+                free(tmp);
+                fprintf(orte_xml_fp, "%s", output);
+                free(output);
+            } else {
+                opal_output(0, "%d more process%s sent help message %s / %s",
+                            tli->tli_count_since_last_display,
+                            (tli->tli_count_since_last_display > 1) ? "es have" : " has",
+                            tli->tli_filename, tli->tli_topic);
+            }
             tli->tli_count_since_last_display = 0;
 
             if (first) {
-                opal_output(0, "Set MCA parameter \"orte_base_help_aggregate\" to 0 to see all help / error messages");
+                if (orte_xml_output) {
+                    fprintf(orte_xml_fp, "<stderr>Set MCA parameter \"orte_base_help_aggregate\" to 0 to see all help / error messages</stderr>\n");
+                    fflush(orte_xml_fp);
+                } else {
+                    opal_output(0, "Set MCA parameter \"orte_base_help_aggregate\" to 0 to see all help / error messages");
+                }
                 first = false;
             }
         }
@@ -273,7 +405,15 @@ static int show_help(const char *filename, const char *topic,
     } 
     /* Not already displayed */
     else if (ORTE_ERR_NOT_FOUND == rc) {
-        fprintf(stderr, "%s", output);
+        if (orte_xml_output) {
+            char *tmp;
+            tmp = xml_format((unsigned char*)output);
+            fprintf(orte_xml_fp, "%s", tmp);
+            fflush(orte_xml_fp);
+            free(tmp);
+        } else {
+            fprintf(stderr, "%s", output);
+        }
         if (!show_help_timer_set) {
             show_help_time_last_displayed = now;
         }
@@ -359,14 +499,17 @@ int orte_show_help_init(void)
 {
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output, "orte_show_help init"));
 
+    /* Show help duplicate detection */
     if (ready) {
         return ORTE_SUCCESS;
     }
     ready = true;
 
-    /* Show help duplicate detection */
     OBJ_CONSTRUCT(&abd_tuples, opal_list_t);
     
+    save_help = opal_show_help;
+    opal_show_help = orte_show_help;
+
     return ORTE_SUCCESS;
 }
 
@@ -376,9 +519,12 @@ void orte_show_help_finalize(void)
         return;
     }
     ready = false;
-    
+
+    opal_show_help = save_help;
+    save_help = NULL;
+
     /* Shutdown show_help, showing final messages */
-    if (orte_process_info.hnp) {
+    if (ORTE_PROC_IS_HNP) {
         show_accumulated_duplicates(0, 0, NULL);
         OBJ_DESTRUCT(&abd_tuples);
         if (show_help_timer_set) {
@@ -389,7 +535,6 @@ void orte_show_help_finalize(void)
         orte_rml.recv_cancel(ORTE_NAME_WILDCARD, ORTE_RML_TAG_SHOW_HELP);
         return;
     }
-
 }
 
 int orte_show_help(const char *filename, const char *topic, 
@@ -424,12 +569,14 @@ int orte_show_help(const char *filename, const char *topic,
     }
     
     /* if we are the HNP, or the RML has not yet been setup,
-     * or we don't yet know our HNP, then all we can do
+     * or ROUTED has not been setup,
+     * or we weren't given an HNP, then all we can do
      * is process this locally
      */
-    if (orte_process_info.hnp ||
+    if (ORTE_PROC_IS_HNP ||
         NULL == orte_rml.send_buffer ||
-        ORTE_PROC_MY_HNP->vpid == ORTE_VPID_INVALID) {
+        NULL == orte_routed.get_route ||
+        NULL == orte_process_info.my_hnp_uri) {
         rc = show_help(filename, topic, output, ORTE_PROC_MY_NAME);
     }
     
